from strongr.clouddomain.handler.abstract.cloud import AbstractRunJobHandler
import sys

from strongr.clouddomain.model.gateways import Gateways

if sys.version_info[0] > 3 and sys.version_info[1] > 3:
    # python > 3.3 uses shlex.quote
    from shlex import quote
else:
    from pipes import quote

import os
import subprocess
import tempfile

import threading
import strongr.core
import time

class RunJobHandler(AbstractRunJobHandler):
    def __call__(self, command):
        thread = threading.Thread(target=self._run, args=(command,)) # run in separate thread so it doesn't block strongr
        thread.start()
        #self._run(command)

    def _run(self, command):
        inter_domain_event_factory = Gateways.inter_domain_event_factory()
        inter_domain_events_publisher = strongr.core.Core.inter_domain_events_publisher()

        config = strongr.core.Core.config()
        scratch_path = config.clouddomain.MockCloud.scratch

        volumes = ''
        env = ''
        if command.scratch:
            if not os.path.isdir(scratch_path):
                os.mkdir(scratch_path, 0700)

            volumes = '--volume={}:/scratch'.format(scratch_path)
            env = "-e SCRATCH_DIR='/scratch'"


        cmd = 'docker run {} {} -di --name {} -m {}g --cpus={} --entrypoint /bin/sh {}'.format(volumes, env, command.job_id, command.memory, command.cores, quote(command.image))
        ret_code = subprocess.call(cmd, shell=True) # start docker container

        print(cmd)

        if ret_code != 0:
            raise Exception('Something went wrong while initializing docker image: {}'.format(cmd))

        time.sleep(1)

        tmpfile = tempfile.mkstemp()[1]
        fh = open(tmpfile, 'w')
        fh.write("\n".join(command.script))
        fh.close()

        cmd = 'docker exec -i {} /bin/sh < {}'.format(command.job_id, tmpfile)
        try:
            stdout = subprocess.check_output(cmd, stderr=subprocess.STDOUT, shell=True)
        except subprocess.CalledProcessError as err:
            job_finished_event = inter_domain_event_factory.newJobFinishedEvent(command.job_id, err.output, err.returncode)
            inter_domain_events_publisher.publish(job_finished_event)
            if ret_code != 0:
                Exception('Something went wrong while executing script in docker image: {}'.format(cmd))
            stdout = err.output

        os.remove(tmpfile)

        time.sleep(1)

        cmd = 'docker stop {}'.format(command.job_id)
        ret_code = subprocess.call(cmd, shell=True)

        if ret_code != 0:
            raise Exception('Something went wrong while stopping docker image: {}'.format(cmd))

        time.sleep(1)

        cmd = 'docker rm {}'.format(command.job_id)
        try:
            subprocess.check_output(cmd, stderr=subprocess.STDOUT, shell=True)
        except subprocess.CalledProcessError as err:
            if ret_code != 0:
                Exception('Something went wrong while executing script in docker image: {}'.format(cmd))
            stdout = err.output

        job_finished_event = inter_domain_event_factory.newJobFinishedEvent(command.job_id, stdout, 0)
        inter_domain_events_publisher.publish(job_finished_event)



from abc import ABCMeta, abstractmethod

class AbstractStats():
    __metaclass__ = ABCMeta

    @abstractmethod
    def incr(self, namespace, amount, rate=None):
        pass

    @abstractmethod
    def decr(self, namespace, amount, rate=None):
        pass

    @abstractmethod
    def timing(self, namespace, timems):
        pass

    @abstractmethod
    def time(self, namespace):
        """
        Should return a context manager
        :param namespace:
        :return:
        """
        pass

    @abstractmethod
    def gauge(self, namespace, amount, rate=False, delta=False):
        pass

    @abstractmethod
    def set(self, namespace, arr):
        pass



from strongr.restdomain.model.gateways import Gateways
from .wrapper import Command

class RunRestServerCommand(Command):
    """
    Runs the strongr REST server that sits between FASTR and STRONGR

    restdomain:startserver
    """
    def handle(self):
        Gateways.flask_app().run()



from .wrapper import Command

import strongr.core.domain.schedulerdomain

class CleanupCommand(Command):
    """
    Run cleanup jobs

    deploy:cleanup
    """
    def handle(self):
        command_factory = strongr.core.domain.schedulerdomain.SchedulerDomain.commandFactory()
        command_bus = strongr.core.domain.schedulerdomain.SchedulerDomain.schedulerService().getCommandBus()

        command_bus.handle(command_factory.newCleanupNodes())



from strongr.core.domain.schedulerdomain import SchedulerDomain
from .wrapper import Command

import strongr.core.domain.schedulerdomain
import json

class GetJobStdOut(Command):
    """
    Returns a jobs stdout

    job:stdout
        {jid : The job id to be checked}
    """
    def handle(self):
        query_factory = strongr.core.domain.schedulerdomain.SchedulerDomain.queryFactory()
        query_bus = strongr.core.domain.schedulerdomain.SchedulerDomain.schedulerService().getQueryBus()

        command = query_factory.newRequestTaskInfo(self.argument('jid'))

        result = query_bus.handle(command)

        if result is None:
            print(json.dumps({}))
            return

        print(json.dumps({'job_id': result.job_id, 'stdout': result.stdout}))



from strongr.configdomain.command import LoadConfig

from strongr.core.exception import InvalidParameterException

class CommandFactory:
    """ this factory instantiates command objects to be sent to a scheduler commandbus. """
    def newLoadConfig(self, environment):
        """ generates a new loadconfig command

        :param environment: the environment that should be loaded
        :type environment: string

        :returns: a loadconfig command object
        :rtype: loadconfig
        """
        if not len(environment) > 0:
            raise InvalidParameterException('environment is invalid')
        return LoadConfig(environment)



from sqlalchemy import and_

import strongr.core.gateways
from strongr.schedulerdomain.model import Job, JobState

class RequestFinishedJobsHandler(object):
    def __call__(self, query, *args, **kwargs):
        session = strongr.core.gateways.Gateways.sqlalchemy_session()

        if query.jobs is not None:
            result = session.query(Job).filter(and_(Job.state.in_([JobState.FAILED, JobState.FINISHED]), Job.job_id.in_(query.jobs))).order_by(
            Job.job_id).all()
        else:
            result = session.query(Job).filter(Job.state.in_([JobState.FAILED, JobState.FINISHED])).order_by(
                Job.job_id).all()

        job_ids = []
        for job in result:
            job_ids.append(job.job_id)
        return job_ids



from strongr.core.exception import InvalidParameterException

from strongr.clouddomain.event.intra.saltjobfinished import SaltJobFinished

# try-except for py 2 / 3 compatibility
try:
    basestring
except NameError:
    basestring = str

class IntraDomainEventFactory(object):
    def newSaltJobFinished(self, jid,  ret, retcode):
        if not isinstance(jid, basestring) or len(jid.strip()) == 0:
            raise InvalidParameterException('jid is invalid')
        elif not isinstance(ret, basestring) or len(ret.strip()) == 0:
            raise InvalidParameterException('ret is invalid')
        elif not isinstance(retcode, int):
            raise InvalidParameterException('retcode is invalid')

        return SaltJobFinished(jid, ret, retcode)



import dependency_injector.containers as containers
import dependency_injector.providers as providers

from strongr.configdomain.service import ConfigService
from strongr.configdomain.factory import CommandFactory

class ConfigDomain(containers.DeclarativeContainer):
    """IoC container of service providers."""
    configService = providers.Singleton(ConfigService)
    commandFactory = providers.Singleton(CommandFactory)



import unittest

import strongr.clouddomain.model.gateways

import strongr.core
import strongr.core.domain.clouddomain

class TestIntraDomainEvent(unittest.TestCase):
    def test_salt_job_finished(self):
        intra_domain_event_factory = strongr.clouddomain.model.gateways.Gateways.intra_domain_event_factory()
        cloudService = strongr.core.domain.clouddomain.CloudDomain.cloudService() # this initializes event subscribers within the domain

        event = intra_domain_event_factory.newSaltJobFinished('1', 'Test return', 0)

        strongr.clouddomain.model.gateways.Gateways.intra_domain_events_publisher().subscribe(event.__class__,
            lambda event: self.assertTrue(True)
        )

        strongr.clouddomain.model.gateways.Gateways.intra_domain_events_publisher().publish(event)



from datetime import datetime, timedelta

from strongr.schedulerdomain.model.scalingdrivers.abstract import AbstractScaleOut, AbstractScaleIn, \
    AbstractVmTemplateRetriever


# The null-scaler ignores scalein and scaleout signals.

class NullScaler(AbstractScaleIn, AbstractScaleOut, AbstractVmTemplateRetriever):
    def __init__(self, config, *args, **kwargs):
        super(NullScaler, self).__init__(*args, **kwargs)
        # ignore config

    def get_vm_max_age(self, vm_name):
        return datetime.utcnow() + timedelta(days=100000)

    def get_templates(self):
        return ['localhost']

    def scalein(self, command):
        pass

    def scaleout(self, command):
        pass




import strongr.core.domain.schedulerdomain
from .wrapper import Command

class EnsureMinAmountOfNodesCommand(Command):
    """
    Ensure the minimum amount of nodes is spawned.

    deploy:environment
    """
    def handle(self):
        schedulerService = strongr.core.domain.schedulerdomain.SchedulerDomain.schedulerService()
        commandFactory = strongr.core.domain.schedulerdomain.SchedulerDomain.commandFactory()

        command = commandFactory.newEnsureMinAmountOfNodes()
        schedulerService.getCommandBus().handle(command)



from strongr.clouddomain.handler.abstract.cloud import AbstractRequestJidStatusHandler

import salt.runner

import strongr.core
import strongr.core.gateways

class RequestJidStatusHandler(AbstractRequestJidStatusHandler):
    def __call__(self, query):
        opts = salt.config.master_config(strongr.core.Core.config().clouddomain.Azure.salt_config + '/master')
        opts['quiet'] = True
        runner = salt.runner.RunnerClient(opts)

        cache = strongr.core.gateways.Gateways.cache()

        if not cache.exists('clouddomain.jobs.running'):
            cache.set('clouddomain.jobs.running', runner.cmd('jobs.active'), 1)

        jobs = cache.get('clouddomain.jobs.running')

        if jobs is None:
            jobs = {}

        if query.jid not in jobs: # we only want to give status when the job is finished running
            result = runner.cmd('jobs.lookup_jid', [query.jid])
            return result
        return None



from strongr.clouddomain.handler.abstract.cloud import AbstractListDeployedVmsHandler

class ListDeployedVmsHandler(AbstractListDeployedVmsHandler):
    def __call__(self, command):
        return {'up': ['localhost'], 'down': []}



from strongr.core.abstracts.abstractservice import AbstractService

from strongr.authdomain.query import IsValidUser
from strongr.authdomain.handler import IsValidUserHandler

class AuthService(AbstractService):
    _command_bus = None
    _query_bus = None

    def getCommandBus(self):
        if self._command_bus is None:
            self._command_bus = self._make_default_commandbus({
                    })
        return self._command_bus

    def getQueryBus(self):
        if self._query_bus is None:
            self._query_bus = self._make_default_querybus({
                    IsValidUserHandler: IsValidUser,
                })
        return self._query_bus



from strongr.clouddomain.handler.abstract.cloud import AbstractRunJobHandler
import strongr.core

import salt.client

class RunJobHandler(AbstractRunJobHandler):
    def __call__(self, command):
        local = salt.client.LocalClient()
        local.cmd_async(command.host, 'cmd.run', [command.script, "runas={}".format(strongr.core.Core.config().clouddomain.Azure.runas)], jid=command.job_id)



from flask import Blueprint
from flask_restplus import Api

from strongr.restdomain.api.v1.scheduler import ns as scheduler_namespace
from strongr.restdomain.api.v1.oauth2 import ns as oauth2_namespace
from strongr.restdomain.api.v1.secrets import ns as secrets_namespace

blueprint = Blueprint('api_v1', __name__, url_prefix='/v1')
api = Api(blueprint,
    title='V1 Api',
    version='1.0',
    description='First version of the API'
)

api.add_namespace(scheduler_namespace)
api.add_namespace(oauth2_namespace)
api.add_namespace(secrets_namespace)



class RunJob:
    def __init__(self, host, image, script, job_id, scratch, cores, memory):
        self.host = host
        self.image = image
        self.script = script
        self.job_id = job_id
        self.scratch = scratch
        self.cores = cores
        self.memory = memory



from strongr.schedulerdomain.model import Job
import strongr.core.gateways

class RequestTaskInfoHandler:
    def __call__(self, query):
        session = strongr.core.gateways.Gateways.sqlalchemy_session()
        result = session.query(Job).filter(Job.job_id.in_(query.jobs)).all()
        return result



from strongr.core import Core
from strongr.core.domain.clouddomain import CloudDomain
from .wrapper import Command

class ListDeployedVmsCommand(Command):
    """
    List VMs deployed in the cloud.

    deploy:list
    """
    def handle(self):
        cloudService = CloudDomain.cloudService()
        queryFactory = CloudDomain.queryFactory()

        queryBus = cloudService.getQueryBus()
        listDeployedVms = queryFactory.newListDeployedVms()
        print(queryBus.handle(listDeployedVms))



from strongr.core import Core
from .wrapper import Command
import json

class PrintConfig(Command):
    """
    Prints the config.

    print:config
    """
    def handle(self):
        config = Core.config().as_dict()
        del config['internal']
        print(json.dumps(config, sort_keys=True))



import strongr.core
from strongr.core.lock.file import FileLock
from strongr.core.lock.redis import RedisLock


def get_lock(*args, **kwargs):
    driver = strongr.core.getCore().config().lock.driver.strip()

    # pass config vars to lock constructor
    if hasattr(strongr.core.getCore().config().lock, driver):
        if kwargs is None:
            kwargs = {}
        kwargs.update(getattr(strongr.core.getCore().config().lock, driver).as_dict())

    if driver == 'redis':
        return RedisLock(*args, **kwargs)
    elif driver == 'file':
        return FileLock(*args, **kwargs)



import strongr.core.gateways
from strongr.schedulerdomain.model import Vm, VmState

class VmCreatedHandler(object):
    def __call__(self, command):
        session = strongr.core.gateways.Gateways.sqlalchemy_session()
        session.query(Vm).filter(Vm.vm_id == command.vm_id).update({Vm.state: VmState.PROVISION}, synchronize_session='fetch')



import strongr.core
from strongr.core.cache import RedisCache, LocalCache


def get_cache(*args, **kwargs):
    driver = strongr.core.getCore().config().cache.driver.strip()

    if driver == 'redis':
        return RedisCache(*args, **kwargs)
    elif driver == 'local':
        return LocalCache(*args, **kwargs)



from .runjobhandler import RunJobHandler
from .listdeployedvmshandler import ListDeployedVmsHandler
from .deployvmshandler import DeployVmsHandler
from .requestjidstatushandler import RequestJidStatusHandler
from .destroyvmshandler import DestroyVmsHandler



from authlib.specs.rfc6749.grants import (
    AuthorizationCodeGrant as _AuthorizationCodeGrant
)
from authlib.common.security import generate_token

from strongr.core.gateways import Gateways
from strongr.restdomain.model.oauth2 import Token
from strongr.restdomain.model.oauth2.authorizationcode import AuthorizationCode


class AuthorizationCode(_AuthorizationCodeGrant):
    def create_authorization_code(self, client, grant_user, **kwargs):
        # you can use other method to generate this code
        code = generate_token(48)
        item = AuthorizationCode(
            code=code,
            client_id=client.client_id,
            redirect_uri=kwargs.get('redirect_uri', ''),
            scope=kwargs.get('scope', ''),
            user_id=grant_user,
        )
        session = Gateways.sqlalchemy_session()
        session.add(item)
        session.commit()
        return code

    def parse_authorization_code(self, code, client):
        item = AuthorizationCode.query.filter_by(
            code=code, client_id=client.client_id).first()
        if item and not item.is_expired():
            return item

    def delete_authorization_code(self, authorization_code):
        session = Gateways.sqlalchemy_session()
        session.delete(authorization_code)
        session.commit()

    def create_access_token(self, token, client, authorization_code):
        item = Token(
            client_id=client.client_id,
            user_id=authorization_code.user_id,
            **token
        )
        session = Gateways.sqlalchemy_session()
        session.add(item)
        session.commit()
        # we can add more data into token
        token['user_id'] = authorization_code.user_id



from strongr.clouddomain.handler.abstract.cloud import AbstractJobFinishedHandler

class JobFinishedHandler(AbstractJobFinishedHandler):
    def __call__(self, command):
        pass
        # convert command to inter-domain event



from strongr.core.exception import IsNotCallableException

import logging

class EventsPublisher:
    def __init__(self, domain):
        self._subscribers = {}
        self._logger = logging.getLogger(domain + self.__class__.__name__)

    def unsubscribe(self, event, subscriber):
        if not callable(subscriber):
            raise IsNotCallableException()

        eventClass = event.__name__
        if eventClass not in self._subscribers:
            return

        self._subscribers[eventClass].remove(subscriber)

    def subscribe(self, event, callable_subscriber):
        eventClass = event.__name__
        if eventClass not in self._subscribers:
            self._subscribers[eventClass] = []

        if not callable(callable_subscriber):
            raise IsNotCallableException()

        self._subscribers[eventClass].append(callable_subscriber)

    def publish(self, event):
        event_class = event.__class__.__name__
        self._logger.debug('publishing event {} {}'.format(event_class, event.__dict__))
        if event_class in self._subscribers:
            for subscriber in self._subscribers[event_class]:
                if subscriber(event):
                    # stop eventClass bubbling on True
                    return False
        return True



import strongr.core.gateways

from strongr.secretsdomain.model import Secret

import itertools

class ListSecretsHandler():
    def __call__(self, query):
        session = strongr.core.gateways.Gateways.sqlalchemy_session()

        result = session.query(Secret.key).order_by(Secret.key).all()

        return list(itertools.chain.from_iterable(result))



import strongr.core.gateways

from strongr.secretsdomain.model import Secret

import itertools

class GetSecretHandler():
    def __call__(self, query):
        session = strongr.core.gateways.Gateways.sqlalchemy_session()

        result = session.query(Secret).filter(Secret.key == query.key).first()

        if result is not None:
            return result.value

        return None



import dependency_injector.containers as containers
import dependency_injector.providers as providers

from strongr.authdomain.service import AuthService
from strongr.authdomain.factory import QueryFactory

class AuthDomain(containers.DeclarativeContainer):
    """IoC container of service providers."""
    authService = providers.Singleton(AuthService)
    commandFactory = providers.Singleton(lambda: None)
    queryFactory = providers.Singleton(QueryFactory)



import strongr.core.gateways

from strongr.schedulerdomain.model import Job, JobState

class JobFinishedHandler(object):
    def __call__(self, command):
        session = strongr.core.gateways.Gateways.sqlalchemy_session()

        jobstate = (JobState.FINISHED if command.retcode == 0 else JobState.FAILED)
        session.query(Job).filter(Job.job_id == command.job_id).update(
            {
                Job.state: jobstate,
                Job.stdout: command.ret,
                Job.return_code: command.retcode
            },
            synchronize_session='fetch'
        )



import pickle
import strongr.core
import strongr.core.gateways
from strongr.core.lock.redis import RedisLock


class RedisCache:
    _redis = None
    _namespace = None

    def __init__(self):
        self._redis = strongr.core.gateways.Gateways.redis()
        self._namespace = strongr.core.Core.config().cache.namespace

    def set(self, key, value, timeout):
        self._redis.setex(self._namespace + key, pickle.dumps(value), timeout)

    def get(self, key):
        ret = self._redis.get(self._namespace + key)
        return None if ret is None else pickle.loads(ret)

    def delete(self, key):
        self._redis.delete(self._namespace + key)

    def exists(self, key):
        return self._redis.exists(self._namespace + key)




from strongr.schedulerdomain.model.scalingdrivers.abstract import AbstractScaleOut, AbstractScaleIn, \
    AbstractVmTemplateRetriever

import uuid

import strongr.core
import strongr.core.gateways
import logging
import strongr.core.domain.schedulerdomain

from sqlalchemy import func, and_, or_

from strongr.schedulerdomain.model import JobState, Job, Vm, VmState

from datetime import datetime, timedelta

# This scaler is specifically written for the Surf-SARA HPCCloud

class SurfHpcScaler(AbstractScaleIn, AbstractScaleOut, AbstractVmTemplateRetriever):
    def __init__(self, config, *args, **kwargs):
        super(SurfHpcScaler, self).__init__(*args, **kwargs)
        self._config = config

    def get_vm_max_age(self, vm_name):
        deadline = datetime.utcnow() + timedelta(hours=6)
        return deadline

    def get_templates(self):
        return self._config['templates']

    def scalein(self, command):
        if strongr.core.gateways.Gateways.lock('scaleout-lock').exists():
            return # only every run one of these commands at once

        # double check lock
        # we should move this locking mechanism a little bit more down the chain so that it isn't the specific scaling
        # algorithms responsibility
        with strongr.core.gateways.Gateways.lock('scaleout-lock'):  # only ever run one of these commands at once
            logger = logging.getLogger('schedulerdomain.' + self.__class__.__name__)

            session = strongr.core.gateways.Gateways.sqlalchemy_session()

            # subquery to see whats already running on vm
            subquery1 = session.query(Job.vm_id, func.count(Job.job_id).label('jobs'), func.sum(Job.cores).label('cores'), func.sum(Job.ram).label('ram')).filter(
                Job.state.in_([JobState.RUNNING])).group_by(Job.vm_id).subquery('j')

            subquery2 = session.query(Job.vm_id, func.max(Job.state_date).label('last_job_date')).filter(Job.state.in_([JobState.FAILED, JobState.FINISHED, JobState.RUNNING])).group_by(Job.vm_id).subquery('i')

            job_deadline = datetime.utcnow() + timedelta(minutes=-10)

            results = session.query(Vm.vm_id.label('vm_id'), subquery1.c.jobs.label('job_count'), subquery2.c.last_job_date) \
                .outerjoin(subquery1, subquery1.c.vm_id == Vm.vm_id) \
                .outerjoin(subquery2, subquery2.c.vm_id == Vm.vm_id) \
                .filter(
                and_(
                    and_(  # only vm's with no jobs
                        subquery1.c.cores == None,
                        subquery1.c.ram == None,
                    ),
                    or_(
                        subquery2.c.last_job_date < job_deadline, # if no job running for 10 minutes kill the vm
                        and_(subquery2.c.last_job_date == None, Vm.state_date < job_deadline) # if no job scheduled to vm for 10 minutes kill the vm
                    ),
                    Vm.state.in_([VmState.READY])  # vm should be in state ready
                )
            ).all()

            if not results:
                return # no VM's to scalein

            vms_to_update = []
            for vm in results:
                vms_to_update.append(vm[0])

            if len(vms_to_update) > 0:
                    session.query(Vm).filter(Vm.vm_id.in_(vms_to_update)).update({Vm.state: VmState.MARKED_FOR_DEATH}, synchronize_session='fetch')

    def scaleout(self, command):
        cores = command.cores
        ram = command.ram

        if strongr.core.gateways.Gateways.lock('scaleout-lock').exists():
            return # only every run one of these commands at once

        with strongr.core.gateways.Gateways.lock('scaleout-lock'):  # only ever run one of these commands at once
            config = self._config
            logger = logging.getLogger('schedulerdomain.' + self.__class__.__name__)

            query_factory = strongr.core.domain.schedulerdomain.SchedulerDomain.queryFactory()
            query_bus = strongr.core.domain.schedulerdomain.SchedulerDomain.schedulerService().getQueryBus()

            templates = dict(config['templates']) # make a copy because we want to manipulate the list

            active_vms = query_bus.handle(query_factory.newRequestVms([VmState.NEW, VmState.PROVISION, VmState.READY]))

            provision_counter = 0
            for vm in active_vms:
                if vm.state in [VmState.NEW, VmState.PROVISION]:
                    cores -= vm.cores
                    ram -= vm.ram
                    provision_counter += 1

            if provision_counter >= 2:
                # don't provision more than 2 VM's at the same time
                return

            for template in templates:
                templates[template]['distance'] = templates[template]['ram'] / templates[template]['cores']

            ram_per_core_needed = ram / cores

            # find best fit based on templates

            # first we calculate the distance based on optimal mem / core distribution
            distances = {}
            for template in templates:
                distance = abs(templates[template]['distance'] - ram_per_core_needed)
                if distance not in distances:
                    distances[distance] = []
                distances[distance].append(template)

            min_distance_templates = distances[min(distances)] # get templates with least distance

            # now find the templates with resources that best fit what we need
            # we simply select the best fitting template with the most resources
            template = min(min_distance_templates, key=lambda key: (cores - templates[key]['cores']) + (ram - templates[key]['ram']))

            # scaleout by one instance
            cloudService = strongr.core.domain.clouddomain.CloudDomain.cloudService()
            cloudCommandFactory = strongr.core.domain.clouddomain.CloudDomain.commandFactory()
            profile = config['default_profile'] if 'profile' not in templates[template] else templates[template]['profile']
            deployVmsCommand = cloudCommandFactory.newDeployVms(names=[template + '-' + str(uuid.uuid4())], profile=profile, cores=templates[template]['cores'], ram=templates[template]['ram'])

            cloudCommandBus = cloudService.getCommandBus()

            logger.info('Deploying VM {0} cores={1} ram={2}GiB'.format(deployVmsCommand.names[0], templates[template]['cores'], templates[template]['ram']))

            cloudCommandBus.handle(deployVmsCommand)



from abc import ABCMeta, abstractmethod

from cmndr import CommandBus
from cmndr.handlers import CommandHandler
from cmndr.handlers.inflectors import CallableInflector
from cmndr.handlers.locators import LazyLoadingInMemoryLocator
from cmndr.handlers.nameextractors import ClassNameExtractor

from strongr.clouddomain.handler.abstract.cloud import AbstractDestroyVmsHandler, AbstractDeployVmsHandler, \
                                                        AbstractListDeployedVmsHandler, AbstractRunJobHandler,\
                                                        AbstractRequestJidStatusHandler, AbstractJobFinishedHandler

from strongr.clouddomain.command import DestroyVms, DeployVms, RunJob, JobFinished
from strongr.clouddomain.query import ListDeployedVms, RequestJidStatus

import strongr.clouddomain.model.gateways as gateways
import strongr.core as core

class AbstractCloudService():
    __metaclass__ = ABCMeta
    _commands = {}
    _queries = {}


    _mappings = {
        AbstractListDeployedVmsHandler: ListDeployedVms.__name__,
        AbstractRunJobHandler: RunJob.__name__,
        AbstractDeployVmsHandler: DeployVms.__name__,
        AbstractRequestJidStatusHandler: RequestJidStatus.__name__,
        AbstractDestroyVmsHandler: DestroyVms.__name__,
        AbstractJobFinishedHandler: JobFinished.__name__
    }

    def __init__(self):
        # map commands and handlers
        for handler in self.get_command_handlers():
            command = self._getCommandForHandler(handler)
            self._commands[handler] = command
        for handler in self.get_query_handlers():
            command = self._getCommandForHandler(handler)
            self._queries[handler] = command

        # map events to commands / queries
        event_bindings = gateways.Gateways.domain_event_bindings()
        for event in event_bindings:
            if 'command' in event_bindings[event]:
                for command_generator in event_bindings[event]['command']:
                    gateways.Gateways.intra_domain_events_publisher().subscribe(event, (lambda event, command_generator=command_generator: self.getCommandBus().handle(command_generator(event))))
            if 'query' in event_bindings[event]:
                for query_generator in event_bindings[event]['query']:
                    gateways.Gateways.intra_domain_events_publisher().subscribe(event, (lambda event, query_generator=query_generator: self.getQueryBus().handle(query_generator(event))))
            if 'escalate-to-inter' in event_bindings[event]: # escalate intra-event to inter-event
                for inter_domain_event_generator in event_bindings[event]['escalate-to-inter']:
                    gateways.Gateways.intra_domain_events_publisher().subscribe(event, (lambda event, inter_domain_event_generator=inter_domain_event_generator: core.Core.inter_domain_events_publisher().publish(inter_domain_event_generator(event))))



    @abstractmethod
    def get_command_handlers(self):
        return

    @abstractmethod
    def get_query_handlers(self):
        pass

    def _getCommandForHandler(self, handler):
        for mappedHandler in self._mappings:
            if issubclass(handler, mappedHandler):
                command = self._mappings[mappedHandler]
                # remove from self._mappings so that a handler with multiple inheritance can work
                del self._mappings[mappedHandler]
                return command
        return None

    def _makeBus(self, handlers, middlewares=None):
        extractor = ClassNameExtractor()
        locator = LazyLoadingInMemoryLocator(handlers)
        inflector = CallableInflector()
        handler = CommandHandler(extractor, locator, inflector)
        if middlewares != None:
            return CommandBus(middlewares + [handler])
        return CommandBus([handler])

    def getCommandBus(self, middlewares=None):
        if not hasattr(self, '_commandbus'):
            self._commandbus = self._makeBus(self._commands, middlewares)
        return self._commandbus

    def getQueryBus(self, middlewares=None):
        if not hasattr(self, '_querybus'):
            self._querybus = self._makeBus(self._queries, middlewares)
        return self._querybus



# from sqlalchemy import func, and_, or_
#
# from strongr.schedulerdomain.model import Job, JobState, VmState, Vm
from strongr.schedulerdomain.model.scalingdrivers.surfhpccloudscaler import SurfHpcScaler
from .wrapper import Command
#
# import strongr.core.gateways

import strongr.core.domain.schedulerdomain

class TestCommand(Command):
    """
    Runs experimental testcode

    test:run
    """
    def handle(self):
        s = SurfHpcScaler(config=dict(strongr.core.Core.config().schedulerdomain.as_dict()[strongr.core.Core.config().schedulerdomain.scalingdriver]))
        s.scaleout(None)



from .cloud import OpenNebula, MockCloud, Azure

# cloud service factory
class CloudServices():
    _clouds = [ \
        OpenNebula,
        MockCloud,
        Azure
    ]

    _instances = {}

    def getCloudNames(self):
        return [cloud.__name__ for cloud in self._clouds]

    def getCloudServiceByName(self, name):
        if name not in self._instances:
            service = next((cloud for cloud in self._clouds if cloud.__name__ == name), None)
            if service == None:
                return None
            self._instances[name] = service()

        return self._instances[name]



from abc import ABCMeta, abstractmethod

import strongr.core
class CallableCommandHandler:
    __metaclass__ = ABCMeta

    @abstractmethod
    def __call__(self, command):
        pass

    def executeAndPublishDomainEvent(self, event, callable, **kwargs):
        callable(*kwargs)
        self.publishDomainEvent(event)

    def publishDomainEvent(self, event):
        core = strongr.core.getCore()
        core.domainEventsPublisher().publish(event)



from cleo import Command

class PatchedCommand(Command):
    def __init__(self):
        super(PatchedCommand, self).__init__()

    def ask(self, question, default):
        # since cleo has a bug that causes it not to return default values we need a wrapper that does exactly that
        output = super(PatchedCommand, self).ask(question)
        if output is None:
            output = default
        return output

    def choice(self, question, options, defaultIndex):
        # for some reason cleo can not handle arrays with 1 el so we need a workaround
        if len(options) > 1:
            output = super(PatchedCommand, self).choice(question, options, defaultIndex)
        else:
            output = options[0]

        return output

    def _castToBool(self, val):
        if isinstance(val, bool):
            return val
        return val.lower() in ['true', '1', 't', 'y', 'yes', 'yeah', 'yup', 'certainly', 'uh-huh', 'yarr']



import dependency_injector.containers as containers
import dependency_injector.providers as providers

from .eventspublisher import EventsPublisher
from strongr.core.middlewares.celery.commandrouter import CommandRouter

class Core(containers.DeclarativeContainer):
    """IoC container of core component providers."""
    config = providers.Configuration('config')

    inter_domain_events_publisher = providers.Singleton(EventsPublisher, 'InterDomain')

    command_router = providers.Singleton(CommandRouter)



from cmndr import Middleware

class StatsMiddleware(Middleware):
    def __init__(self, stats):
        self._stats = stats

    def execute(self, command, next_callable):
        namespace = command.__module__.split('.')[1] + '.' + command.__class__.__name__.lower()
        stats = self._stats

        stats.incr(namespace, 1)
        with stats.time(namespace):
            ret = next_callable(command)

        return ret



from sqlalchemy.orm import relationship

import strongr.core.gateways as gateways
from sqlalchemy import Column, ForeignKey, Integer, String

from authlib.flask.oauth2.sqla import OAuth2ClientMixin

from strongr.core.sqlalchemydatatypes.uuidtype import UUIDType

Base = gateways.Gateways.sqlalchemy_base()

class Client(Base, OAuth2ClientMixin):
    __tablename__ = 'oauth_client'

    id = Column(Integer, primary_key=True) # client_id is provided by authlib, can't use it here
    user_id = Column(String(64), ForeignKey('oauth_users.user_id', ondelete='CASCADE'))

    @classmethod
    def get_by_client_id(cls, client_id): # this method in OAuth2ClientMixin is bugged in authlib, thus we must patch it
        client = gateways.Gateways.sqlalchemy_session().query(Client).filter_by(client_id=client_id).first()
        return client




class DefaultsLoader:
    def getConfig(self):
        return {
            'internal': {
                'configloaderorder': ['IniLoader', 'JsonLoader', 'YamlLoader']
            },
            'cache': {
                'driver': 'local',
                'namespace': 'strongr-cache-'
            },
            'lock': {
                'driver': 'local',
                'redis': {
                    'timeout': 10,
                    'namespace': 'strongr-redislock-'
                },
                'file': {
                    'timeout': 10,
                    'path': '/var/lock/strongr'
                }
            },
            'db': {
                'engine': {
                    'url': 'sqlite:////tmp/strongr.db',
                    'echo': True
                }
            },
            'stats': {
                'driver': "null"
            },
            'clouddomain': {
                'driver': 'MockCloud',
                'OpenNebula': {
                    'salt_config': '/etc/salt'
                },
                'MockCloud': {
                    'scratch': '/tmp/strongr_mockcloud_scratch'
                }
            },
            'restdomain': {
                'backend': 'flask',
                'flask': {
                    'host': '127.0.0.1',
                    'port': 8080,
                    'debug': True
                },
                "gunicorn": {
                    "bind": "0.0.0.0:8080",
                    "worker_class": "sync"
                }
            },
            'schedulerdomain': {
                'scalingdriver': 'nullscaler',
                'simplescaler': {
                    'scaleoutmincoresneeded': 1,
                    'scaleoutminramneeded': 1
                }
            },
            'logger': {
                'version': 1,
                'disable_existing_loggers': False,
                'formatters': {
                    'standard': {
                        'format': '%(asctime)s [%(levelname)s] %(name)s: %(message)s'
                    }
                },
                'handlers': {
                    'default': {
                        'level': 'INFO',
                        'formatter': 'standard',
                        'class': 'logging.StreamHandler'
                    }
                },
                'loggers': {
                    '': {
                        'handlers': [
                            'default'
                        ],
                        'level': 'INFO',
                        'propagate': True
                    }
                }
            }
        }



from strongr.clouddomain.handler.impl.cloud.opennebula import ListDeployedVmsHandler, \
    RunJobHandler, DeployVmsHandler, \
    RequestJidStatusHandler, DestroyVmsHandler, \
    JobFinishedHandler

from .abstractcloudservice import AbstractCloudService

import strongr.clouddomain.model.gateways


class OpenNebula(AbstractCloudService):
    _salt_event_translator_thread = None

    def __init__(self, *args, **kwargs):
        super(OpenNebula, self).__init__(*args, **kwargs)

    def start_reactor(self):
        salt_event_translator = strongr.clouddomain.model.gateways.Gateways.salt_event_translator()
        if not salt_event_translator.is_alive():
            salt_event_translator.setDaemon(True)
            salt_event_translator.start()  # start event translator thread if it wasn't running

    def get_command_handlers(self):
        return [RunJobHandler, DeployVmsHandler, DestroyVmsHandler, JobFinishedHandler]

    def get_query_handlers(self):
        return [ListDeployedVmsHandler, RequestJidStatusHandler]



from strongr.clouddomain.handler.abstract.cloud import AbstractDestroyVmsHandler

class DestroyVmsHandler(AbstractDestroyVmsHandler):
    def __call__(self, command):
        pass # we cant deploy vms in the mockcloud



from strongr.clouddomain.command import DeployVms, RunJob, DestroyVms, JobFinished

from strongr.core.exception import InvalidParameterException

class CommandFactory:
    """ This factory instantiates command objects to be sent to a cloud commandbus. """

    def newJobFinishedCommand(self, job_id, ret, retcode):
        """ Generates a new JobFinished command

        :param job_id: An identifier token for the job
        :type job_id: string

        :param ret: Usually the stdout of the job
        :type ret: string

        :param retcode: The exit code of the job
        :type ret: int

        :returns: A JobFinished command object
        :rtype: JobFinished
        """
        if not isinstance(job_id, basestring) or len(job_id.strip()) == 0:
            raise InvalidParameterException('jid is invalid')
        elif not isinstance(ret, basestring) or len(ret.strip()) == 0:
            raise InvalidParameterException('ret is invalid')
        elif not isinstance(retcode, int):
            raise InvalidParameterException('retcode is invalid')

        return JobFinished(job_id, ret, retcode)

    def newDestroyVms(self, names):
        """ Generates a new DestroyVm command

        :param name: The name of the VM to be destroyed
        :type name: string

        :returns: A DestroyVm command object
        :rtype: DestroyVm
        """
        if not isinstance(names, list) or len(names) <= 0:
            raise InvalidParameterException('names is invalid')

        return DestroyVms(names=names)

    def newDeployVms(self, names, profile, cores, ram):
        """ Generates a new DeployVms command

        :param names: A list of names
        :type names: list

        :param profile: the vm profile to be used
        :type profile: string

        :param cores: the number of cores per vm
        :type cores: int

        :param ram: the amount of ram per vm in GiB
        :type ram: int

        :returns: A DeployVms command object
        :rtype: DeployVms
        """

        if not isinstance(names, list) or len(names) <= 0:
            raise InvalidParameterException('names is inavlid')

        if not isinstance(cores, int) or cores <= 0:
            raise InvalidParameterException('cores is inavlid')

        if not isinstance(ram, int) or ram <= 0:
            raise InvalidParameterException('ram is invalid')

        if len(profile) <= 0:
            raise InvalidParameterException('profile is invalid')

        return DeployVms(names, profile, cores, ram)

    def newRunJob(self, host, image, script, job_id, scratch, cores, memory):
        """ Generates a new RunShellCode command

        :param host: where host where te command should be ran
        :type host: string
        :param image: the docker image the script should run under
        :type image: string
        :param script: array of strings, the shellcode to be ran in the docker container
        :type script: list
        :param job_id: the name of the job to be used
        :type job_id: string
        :param scratch: should a scratch be mounted?
        :type scratch: bool
        :param cores: how many cores for this job?
        :type cores: int
        :param memory: how much memory for this job?
        :type memory: int

        :returns: A new RunJob command object
        :rtype: RunJob
        """

        if not len(host) > 0:
            raise InvalidParameterException('host is invalid')
        elif not len(image) > 0:
            raise InvalidParameterException('image is invalid')
        elif not len(script) > 0:
            raise InvalidParameterException('script is invalid')
        elif not len(job_id) > 0:
            raise InvalidParameterException('job_id is invalid')
        elif cores <= 0:
            raise InvalidParameterException('cores is invalid')
        elif memory <= 0:
            raise InvalidParameterException('memory is invalid')

        return RunJob(host=host, image=image, script=script, job_id=job_id, scratch=scratch, cores=cores, memory=memory)



from strongr.secretsdomain.query import ListSecrets, GetSecret

class QueryFactory():
    _list_secrets = None

    def new_list_secrets(self):
        # flyweight pattern
        if self._list_secrets is None:
            self._list_secrets = ListSecrets()

        return self._list_secrets

    def new_get_secret(self, key):
        return GetSecret(key)



from strongr.clouddomain.handler.abstract.cloud import AbstractDeployVmsHandler

class DeployVmsHandler(AbstractDeployVmsHandler):
    def __call__(self, command):
        pass # we can't deploy VM's in the mockcloud



from authlib.specs.rfc6749.grants import (
    ResourceOwnerPasswordCredentialsGrant as _PasswordGrant
)

from strongr.core.gateways import Gateways
from strongr.restdomain.model.oauth2 import User, Token


class PasswordGrant(_PasswordGrant):
    def authenticate_user(self, username, password):
        user = User.query.filter_by(username=username).first()
        if user.check_password(password):
            return user

    def create_access_token(self, token, client, user, **kwargs):
        item = Token(
            client_id=client.client_id,
            user_id=user.user_id,
            **token
        )
        session = Gateways.sqlalchemy_session()
        session.add(item)
        session.commit()



import dependency_injector.containers as containers
import dependency_injector.providers as providers

from strongr.secretsdomain.service import SecretService

import dependency_injector.providers as providers

from strongr.secretsdomain.service import SecretService
from strongr.secretsdomain.factory import CommandFactory, QueryFactory

import strongr.core.domain.clouddomain


class SecretDomain(containers.DeclarativeContainer):
    """IoC container of service providers."""
    secret_service = providers.Singleton(SecretService)
    command_factory = providers.Singleton(CommandFactory)
    query_factory = providers.Singleton(QueryFactory)



import strongr.core.gateways
from strongr.schedulerdomain.model import Job, JobState

class RequestScheduledTasksHandler:
    def __call__(self, query):
        session = strongr.core.gateways.Gateways.sqlalchemy_session()
        result = session.query(Job).filter(Job.state.in_([JobState.ASSIGNED, JobState.ENQUEUED])).order_by(Job.job_id).limit(query.limit).all()
        return result



from strongr.core.domain.authdomain import AuthDomain
from .wrapper import Command

import json

class AddClient(Command):
    """
    Check if the login is a valid user

    authdomain:addclient
    """
    def handle(self):
        authService = AuthDomain.authService()
        #queryFactory = AuthDomain.queryFactory()

        query = queryFactory.newIsValidUserQuery(self.argument('username'), self.argument('password'))
        result = authService.getQueryBus().handle(query)
        print(json.dumps(result))



from strongr.core import Core
from strongr.core.domain.clouddomain import CloudDomain
from .wrapper import Command

import uuid

class DeployManyCommand(Command):
    """
    Deploys VM's in the cloud in a parallel way. A first step towards elasticity.

    deploy:many
    """
    def handle(self):
        cloudService = CloudDomain.cloudService()
        commandFactory = CloudDomain.commandFactory()

        cores = int(self.ask('How many processing cores should the VM\'s have? (default 1): ', 1))
        ram = int(self.ask('How much memory in GiB should the VM\'s have? (default 4): ', 4))
        amount = int(self.ask('How many VM\'s should be deployed? (default 2)', 2))

        if not (cores > 0 and ram > 0 and amount > 0):
            # TODO: put something sensible in here, this is just a placeholder
            self.error('Invalid input')
            return

        deployVmNameList = []
        while amount > 0:
            deployVmNameList.append('worker-' + str(uuid.uuid4()))
            amount -= 1

        scaling_driver = Core.config().schedulerdomain.scalingdriver

        profile = getattr(Core.config().schedulerdomain, scaling_driver).default_profile
        deployVms = commandFactory.newDeployVms(names=deployVmNameList, profile=profile, cores=cores, ram=ram)
        commandBus = cloudService.getCommandBus()

        self.info('Deploying {0} VM\'s with cores={1} ram={2}GiB'.format(len(deployVmNameList), cores, ram))

        commandBus.handle(deployVms)



from strongr.clouddomain.handler.abstract.cloud import AbstractListDeployedVmsHandler

import salt.cloud
import strongr.core

class ListDeployedVmsHandler(AbstractListDeployedVmsHandler):
    def __call__(self, command):
        client = salt.cloud.CloudClient(strongr.core.Core.config().clouddomain.OpenNebula.salt_config + '/cloud')
        names = {}
        rs = client.query()

        for provider in list(rs.keys()):
            for location in list(rs[provider].keys()):
                for machine in list(rs[provider][location].keys()):
                    names[machine] = {
                        'cores': int(rs[provider][location][machine]['size']['cpu']),
                        'ram': int(rs[provider][location][machine]['size']['memory']) // 1024
                    }

        opts = salt.config.master_config(strongr.core.Core.config().clouddomain.OpenNebula.salt_config + '/master')
        opts['quiet'] = True
        runner = salt.runner.RunnerClient(opts)

        result = runner.cmd('manage.up')

        output = {'up': [], 'down': []}
        for machine in list(names):
            if machine not in result:
                output['down'].append(machine)
            else:
                output['up'].append(machine)

        return output



import strongr.core.gateways
from strongr.schedulerdomain.model import Vm, VmState

class VmNewHandler(object):
    def __call__(self, command):
        session = strongr.core.gateways.Gateways.sqlalchemy_session()
        vm = Vm()
        vm.vm_id = command.vm_id
        vm.state = VmState.NEW
        vm.cores = command.cores
        vm.ram = command.ram

        session.add(vm)



class AppendGrant():
    def __init__(self, client_id, code, redirect_uri, scope, user_id, expires):
        self.client_id = client_id
        self.code = code
        self.redirect_uri = redirect_uri
        self.scope = scope
        self.user_id = user_id
        self.expires = expires



from strongr.configdomain.model import DefaultsLoader, IniLoader, JsonLoader, YamlLoader

import strongr.core

class LoadConfigHandler():
    def __call__(self, command):
        config_dict = DefaultsLoader().getConfig()
        strongr.core.getCore().config.update(ConfigStruct(**config_dict))

        loaders = {
            IniLoader.__name__: IniLoader(),
            JsonLoader.__name__: JsonLoader(),
            YamlLoader.__name__: YamlLoader()
        }

        loadOrder = ConfigStruct(**config_dict).internal.configloaderorder
        for loaderName in loadOrder:
            if loaderName in loaders:
                config_dict = self._deep_update(config_dict, loaders[loaderName].getConfig(command.environment))
                # update config in core many times as config loaders can use config parameters from other loaders
                # example: a database configloader that needs credentials
                strongr.core.getCore().config.update(ConfigStruct(**config_dict))
            else:
                # fatal error
                # since the loadOrder is hardcoded in the defaultsloader it should always work. If not,
                # it is considered a fatal error.
                raise Exception("Invalid config loader!")

    def _deep_update(self, dict_left, dict_right):
        for key, value in dict_right.items():
            if isinstance(value, dict):
                result = self._deep_update(dict_left.get(key, {}), value)
                dict_left[key] = result
            else:
                dict_left[key] = dict_right[key]
        return dict_left


class ConfigStruct(object):
    def __init__(self, **entries):
        for key in entries:
            if isinstance(entries[key], dict):
                entries[key] = ConfigStruct(**entries[key])
        self.__dict__.update(entries)

    def as_dict(self):
        output = self.__dict__
        for key in output:
            if isinstance(output[key], ConfigStruct):
                output[key] = output[key].as_dict()
        return output



import strongr.core.gateways as gateways
from sqlalchemy import Column, ForeignKey, Integer, String, Enum, DateTime, func, LargeBinary, Text, Boolean
from sqlalchemy.orm import relationship, synonym

from strongr.schedulerdomain.model import JobState

Base = gateways.Gateways.sqlalchemy_base()

class Job(Base):
    __tablename__ = 'jobs'

    job_id = Column(String(64), primary_key=True)
    cores = Column(Integer)
    ram = Column(Integer)
    script = Column(Text)
    image = Column(Text)
    scratch = Column(Boolean)

    _secrets = Column(Text)

    vm_id = Column(String(255), ForeignKey('vms.vm_id'))
    vm = relationship('Vm', back_populates='jobs')

    return_code = Column(Integer)

    stdout = Column('stdout', LargeBinary(length=52428800)) # we should update this to point to a file at some point as it will bloat the database pretty quickly

    _state = Column('state', Enum(JobState))

    @property
    def secrets(self):
        return self._secrets.split('\n')

    # update state_date-field as well when we update state-field
    @secrets.setter
    def secrets(self, array_of_secrets):
        self._secrets = '\n'.join(array_of_secrets)

    # create a synonym so that _state and state are considered the same field by the mapper
    secrets = synonym('_secrets', descriptor=secrets)

    # In classical SQL we would put a trigger to update this field with NOW() if the state-field is updated.
    # SQLAlchemy has no way to write triggers without writing platform-dependent SQL at the time of writing.
    # Instead we use a setter on the state-field, this setter updates the state_date as well.
    # The disadvantage of this solution is that other apps need to implement logic like this as well making
    # the solution less portable.
    state_date = Column(DateTime())

    @property
    def state(self):
        return self._state

    # update state_date-field as well when we update state-field
    @state.setter
    def state(self, value):
        self._state = value
        self.state_date = func.now()

    # create a synonym so that _state and state are considered the same field by the mapper
    state = synonym('_state', descriptor=state)



    # # use zlib / gzip compression for stdout
    # @property
    # def stdout(self):
    #     return zlib.decompress(self._stdout)
    #
    # @stdout.setter
    # def stdout(self, value):
    #     self._stdout = zlib.compress(value, 9)
    #
    # # create a synonym so that _stdout and stdout are considered the same field by the mapper
    # stdout = synonym('_stdout', descriptor=stdout)



from strongr.core import Core
from strongr.core.domain.clouddomain import CloudDomain
from .wrapper import Command

class DestroyManyCommand(Command):
    """
    Destroys a list of VMs in the cloud.

    destroy:many
        {machines* : The names of the VMs to be destroyed}
    """
    def handle(self):
        cloudService = CloudDomain.cloudService()
        commandFactory = CloudDomain.commandFactory()

        machines = self.argument('machines')

        destroyVmsCommand = commandFactory.newDestroyVms(names=machines)

        commandBus = cloudService.getCommandBus()

        self.info('Destroying VMs {0}'.format(machines))

        commandBus.handle(destroyVmsCommand)



from strongr.clouddomain.handler.abstract.cloud import AbstractRequestJidStatusHandler

class RequestJidStatusHandler(AbstractRequestJidStatusHandler):
    def __call__(self, query):
        pass # not really required yet



from .deploysinglecommand import DeploySingleCommand
from .listdeployedvmscommand import ListDeployedVmsCommand
from .deploymanycommand import DeployManyCommand
from .getfinishedjobscommand import GetFinishedJobsCommand
from .runresourcemanager import RunResourceManager
from .printconfig import PrintConfig
from .isvalidusercommand import IsValidUserCommand
from .runrestservercommand import RunRestServerCommand
from .runworkercommand import RunWorkerCommand
from .destroysinglecommand import DestroySingleCommand
from .destroymanycommand import DestroyManyCommand
from .ensureminamountofnodescommand import EnsureMinAmountOfNodesCommand
from .makedbcommand import MakeDbCommand
from .runtestscommand import RunTestsCommand
from .testcommand import TestCommand
from .cleanupcommand import CleanupCommand
from .getjobstdout import GetJobStdOut
from .getsecret import GetSecret



from .appendgranthandler import AppendGrantHandler
from .retrieveclienthandler import RetrieveClientHandler
from .retrievetokenbyaccesstokenhandler import RetrieveTokenByAccessTokenHandler
from .retrievetokenbyrefreshtokenhandler import RetrieveTokenByRefreshTokenHandler
from .retrievegranthandler import RetrieveGrantHandler



from sqlalchemy.types import TypeDecorator, CHAR

from sqlalchemy.dialects.postgresql import UUID as POSTGRESQL_UUID
from sqlalchemy.dialects.mysql import BINARY as MYSQL_BINARY
import uuid


# source: http://docs.sqlalchemy.org/en/latest/core/custom_types.html

class UUIDType(TypeDecorator):
    """Platform-independent GUID type.

    Uses PostgreSQL's UUID type or mysql's binary type, otherwise uses
    CHAR(32), storing as stringified hex values.

    """
    impl = CHAR

    def load_dialect_impl(self, dialect):
        if dialect.name == 'postgresql':
            return dialect.type_descriptor(POSTGRESQL_UUID())
        elif dialect.name == 'mysql':
            return dialect.type_descriptor(MYSQL_BINARY(16))
        else:
            return dialect.type_descriptor(CHAR(32))

    def process_bind_param(self, value, dialect):
        if value is None:
            return value
        elif dialect.name == 'postgresql':
            return str(value)
        elif dialect.name == 'mysql':
            if not isinstance(value, uuid.UUID):
                return uuid.UUID(value).bytes
            else:
                return value.bytes
        else:
            if not isinstance(value, uuid.UUID):
                return "%.32x" % uuid.UUID(value).int
            else:
                # hexstring
                return "%.32x" % value.int

    def process_result_value(self, value, dialect):
        if value is None:
            return value
        elif dialect.name == 'mysql':
            if not isinstance(value, uuid.UUID):
                value = uuid.UUID(bytes=value)
            return value
        else:
            if not isinstance(value, uuid.UUID):
                value = uuid.UUID(value)
            return value



from strongr.schedulerdomain.command import ScheduleJob, RunEnqueuedJobs,\
                                            StartJobOnVm, CheckJobsRunning,\
                                            EnsureMinAmountOfNodes, ScaleOut,\
                                            JobFinished, VmDestroyed,\
                                            VmReady, VmCreated, VmNew, CheckScaling,\
                                            CleanupNodes, ScaleIn, LogStats, CleanupOldJobs

from strongr.core.exception import InvalidParameterException

# try-except for py 2 / 3 compatibility
try:
    basestring
except NameError:
    basestring = str

class CommandFactory:
    """ This factory instantiates command objects to be sent to a scheduler commandbus. """

    def newCleanupOldJobs(self):
        return CleanupOldJobs()

    def newLogStats(self):
        """Generates a new LogStats command"""
        return LogStats()

    def newCleanupNodes(self):
        """ Generates a new Cleanupnodes command

        :return: A CleanupNodes command object
        :rtype: CleanupNodes
        """
        return CleanupNodes()

    def newCheckScaling(self):
        """ Generates a new CheckScaling command

        :returns: A CheckScaling command object
        :rtype: CheckScaling
        """
        return CheckScaling()

    def newVmNew(self, vm_id, cores, ram):
        """ Generates a new VmNew command

        :param vm_id: An identifier token for the job
        :type vm_id: string

        :param cores: the amount of cores assigned to the vm
        :type cores: int

        :param ram: the amount of ram in GiB assigned to the vm
        :type ram: int

        :returns: A VmNew command object
        :rtype: VmNew
        """
        if not isinstance(vm_id, basestring) or len(vm_id.strip()) == 0:
            raise InvalidParameterException('vm_id is invalid')
        elif not isinstance(cores, int) or cores <= 0:
            raise InvalidParameterException('cores is invalid')
        elif not isinstance(ram, int) or ram <= 0:
            raise InvalidParameterException('ram is invalid')

        return VmNew(vm_id, cores, ram)

    def newVmReady(self, vm_id):
        """ Generates a new VmReady command

        :param vm_id: An identifier token for the job
        :type job_id: string

        :returns: A VmReady command object
        :rtype: VmReady
        """
        if not isinstance(vm_id, basestring) or len(vm_id.strip()) == 0:
            raise InvalidParameterException('job_id is invalid')
        return VmReady(vm_id)

    def newVmDestroyed(self, vm_id):
        """ Generates a new VmDestroyed command

        :param vm_id: An identifier token for the job
        :type vm_id: string

        :returns: A VmDestroyed command object
        :rtype: VmDestroyed
        """
        if not isinstance(vm_id, basestring) or len(vm_id.strip()) == 0:
            raise InvalidParameterException('vm_id is invalid')
        return VmDestroyed(vm_id)

    def newVmCreated(self, vm_id):
        """ Generates a new VmCreated command

        :param vm_id: An identifier token for the job
        :type vm_id: string

        :returns: A VmCreated command object
        :rtype: VmCreated
        """
        if not isinstance(vm_id, basestring) or len(vm_id.strip()) == 0:
            raise InvalidParameterException('vm_id is invalid')
        return VmCreated(vm_id)

    def newJobFinished(self, job_id, ret, retcode):
        """ Generates a new JobFinished command

        :param job_id: An identifier token for the job
        :type job_id: string

        :param ret: Usually the stdout of the job
        :type ret: string

        :param retcode: The exit code of the job
        :type ret: int

        :returns: A JobFinished command object
        :rtype: JobFinished
        """
        if not isinstance(job_id, basestring) or len(job_id.strip()) == 0:
            raise InvalidParameterException('jid is invalid')
        elif not isinstance(ret, basestring) or len(ret.strip()) == 0:
            raise InvalidParameterException('ret is invalid')
        elif not isinstance(retcode, int):
            raise InvalidParameterException('retcode is invalid')

        return JobFinished(job_id, ret, retcode)

    def newScaleIn(self):
        return ScaleIn()

    def newScaleOut(self, cores, ram):
        if not cores > 0:
            raise InvalidParameterException('Cores should be higher than 0')
        elif not ram > 0:
            raise InvalidParameterException('Ram should be higher than 0')

        return ScaleOut(cores, ram)

    def newEnsureMinAmountOfNodes(self):
        return EnsureMinAmountOfNodes()

    def newRunEnqueuedJobs(self):
        """ Generates a new RunEnqueuedJobs command

        :returns: A RunEnqueuedJobs command object
        :rtype: RunEnqueuedJobs
        """
        return RunEnqueuedJobs()

    def newScheduleJobCommand(self, image, script, job_id, scratch, cores, memory, secrets):
        """ Generates a new ScheduleTask command

        :param image: the docker image used for running the job
        :type image: string

        :param script: a list of strings with shellcode
        :type script: list

        :param job_id: the jobs id (max 32 characters)
        :type job_id: string

        :param scratch: should a scratch volume be mounted?
        :type scratch: bool

        :param cores: The amount of cores in the VM
        :type cores: int

        :param ram: The amount of RAM in GiB in the VM
        :type ram: int

        :param secrets: List of keys of secrets that should be injected into the process
        :type secrets: list

        :returns: A ScheduleJob command object
        :rtype: ScheduleJob
        """

        if not len(image) > 0:
            raise InvalidParameterException('image is invalid')
        elif not len(script) > 0:
            raise InvalidParameterException('script is invalid')
        elif not len(job_id) > 0:
            raise InvalidParameterException('job_id is invalid')
        elif cores <= 0:
            raise InvalidParameterException('cores is invalid')
        elif memory <= 0:
            raise InvalidParameterException('memory is invalid')
        elif not isinstance(secrets, list):
            raise InvalidParameterException('List of secrets is invalid')

        return ScheduleJob(image=image, script=script, job_id=job_id, scratch=scratch, cores=cores, memory=memory, secrets=secrets)

    def newStartJobOnVm(self, vm_id, job_id):
        """ Generates a new StartJobOnVm command

        :param vm_id: the node name
        :type vm_id: string
        :param job_id: the taskid to be started
        :type job_id: string

        :returns: A StartJobOnVm command object
        :rtype: StartJobOnVm
        """
        if not len(vm_id) > 0:
            raise InvalidParameterException('node is invalid')
        elif not len(job_id) > 0:
            raise InvalidParameterException('taskid is invalid')

        return StartJobOnVm(vm_id=vm_id, job_id=job_id)

    def newCheckJobsRunning(self):
        """ Generates a new CheckJobsRunning command

        :returns: A CheckJobRunning command object
        :rtype: CheckJobsRunning
        """

        return CheckJobsRunning()



import strongr.core.gateways
from strongr.schedulerdomain.model import Job, JobState


class ScheduleJobHandler:
    def __call__(self, command):
        job = Job()
        job.image = command.image
        job.script = "\n".join(command.script)
        job.job_id = command.job_id
        job.scratch = command.scratch
        job.cores = command.cores
        job.ram = command.memory
        job.state = JobState.ENQUEUED
        job.vm = None

        session = strongr.core.gateways.Gateways.sqlalchemy_session()
        session.add(job)
        session.commit()



import dependency_injector.containers as containers
import dependency_injector.providers as providers

import strongr.core.eventspublisher
import strongr.clouddomain.model.salt.salteventtranslator

import strongr.clouddomain.factory.intradomaineventfactory as intradomaineventfactory
import strongr.clouddomain.factory.interdomaineventfactory as interdomainfactory

import strongr.clouddomain.event.intra.saltjobfinished as saltjobfinished

import strongr.core.domain.clouddomain


class Gateways(containers.DeclarativeContainer):
    inter_domain_event_factory = providers.Singleton(interdomainfactory.InterDomainEventFactory)
    intra_domain_event_factory = providers.Singleton(intradomaineventfactory.IntraDomainEventFactory)

    intra_domain_events_publisher = providers.Singleton(strongr.core.eventspublisher.EventsPublisher, 'CloudDomain')
    domain_event_bindings = providers.Object({
        saltjobfinished.SaltJobFinished: { # command / query generators based on event
            #'command': [
            #    #(lambda event: strongr.core.domain.clouddomain.CloudDomain.commandFactory().newJobFinishedCommand(event.jid, event.ret, event.retcode))
            #],
            #'escalate-to-inter': [ # escalate event to inter-domain event
            #    #(lambda event: Gateways.inter_domain_event_factory().newJobFinishedEvent(event.jid, event.ret, event.retcode))
            #]
        }
    })
    salt_event_translator = providers.Singleton(strongr.clouddomain.model.salt.salteventtranslator.SaltEventTranslator, name="SaltEventTranslator")



from strongr.authdomain.query import IsValidUser

from strongr.core.exception import InvalidParameterException

class QueryFactory:
    def newIsValidUserQuery(self, username, password):
        """ Generates a new IsValidUser query

        :returns: An IsValidUser query object
        :rtype: IsValidUser
        """
        if username == None or len(username) == 0 or len(username.strip()) == 0:
            raise InvalidParameterException('Username is invalid')

        if password == None or len(password) == 0:
            raise InvalidParameterException('Password is invalid')

        return IsValidUser(username, password)



from strongr.core.abstracts.abstractservice import AbstractService

from strongr.schedulerdomain.command import ScheduleJob, RunEnqueuedJobs,\
                                            StartJobOnVm, CheckJobsRunning, \
                                            EnsureMinAmountOfNodes, ScaleOut, \
                                            JobFinished, VmCreated,\
                                            VmReady, VmDestroyed, VmNew, CheckScaling,\
                                            CleanupNodes, ScaleIn, LogStats, CleanupOldJobs

from strongr.schedulerdomain.handler import ScheduleJobHandler, RunEnqueuedJobsHandler,\
                                            StartJobOnVmHandler, CheckJobsRunningHandler,\
                                            EnsureMinAmountOfNodesHandler, ScaleOutHandler, \
                                            RequestFinishedJobsHandler, JobFinishedHandler,\
                                            VmDestroyedHandler, VmReadyHandler,\
                                            VmCreatedHandler, VmNewHandler, CheckScalingHandler,\
                                            CleanupNodesHandler, ScaleInHandler, LogStatsHandler,\
                                            CleanupOldJobsHandler

from strongr.schedulerdomain.query import RequestScheduledJobs, RequestJobInfo,\
                                            FindNodeWithAvailableResources, RequestFinishedJobs,\
                                            RequestResourcesRequired, RequestVmsByState

from strongr.schedulerdomain.handler import RequestScheduledTasksHandler, RequestTaskInfoHandler,\
                                            FindNodeWithAvailableResourcesHandler, RequestResourcesRequiredHandler,\
                                            RequestVmsByStateHandler

class SchedulerService(AbstractService):
    _command_bus = None
    _query_bus = None

    def register_models(self):
        import strongr.schedulerdomain.model
        # importing alone is enough for registration

    def getCommandBus(self):
        if self._command_bus is None:
            self._command_bus = self._make_default_commandbus({
                        ScheduleJobHandler: ScheduleJob,
                        RunEnqueuedJobsHandler: RunEnqueuedJobs,
                        StartJobOnVmHandler: StartJobOnVm,
                        CheckJobsRunningHandler: CheckJobsRunning,
                        EnsureMinAmountOfNodesHandler: EnsureMinAmountOfNodes,
                        ScaleOutHandler: ScaleOut,
                        JobFinishedHandler: JobFinished,
                        VmCreatedHandler: VmCreated,
                        VmDestroyedHandler: VmDestroyed,
                        VmReadyHandler: VmReady,
                        VmNewHandler: VmNew,
                        CheckScalingHandler: CheckScaling,
                        CleanupNodesHandler: CleanupNodes,
                        ScaleInHandler: ScaleIn,
                        LogStatsHandler: LogStats,
                        CleanupOldJobsHandler: CleanupOldJobs
                    })
        return self._command_bus

    def getQueryBus(self):
        if self._query_bus is None:
            self._query_bus = self._make_default_querybus({
                    RequestScheduledTasksHandler: RequestScheduledJobs,
                    RequestTaskInfoHandler: RequestJobInfo,
                    FindNodeWithAvailableResourcesHandler: FindNodeWithAvailableResources,
                    RequestFinishedJobsHandler: RequestFinishedJobs,
                    RequestResourcesRequiredHandler: RequestResourcesRequired,
                    RequestVmsByStateHandler: RequestVmsByState
                })
        return self._query_bus



from strongr.core.gateways import Gateways
from strongr.core.domain.schedulerdomain import SchedulerDomain
from strongr.core.domain.restdomain import RestDomain
from strongr.core.domain.secretdomain import SecretDomain

from .wrapper import Command

class MakeDbCommand(Command):
    """
    Create empty database

    database:create
    """
    def handle(self):
        services = [
            SchedulerDomain.schedulerService(),
            RestDomain.oauth2Service(),
            SecretDomain.secret_service()
        ]

        for service in services:
            service.register_models()


        db = Gateways.sqlalchemy_engine()
        metadata = Gateways.sqlalchemy_base().metadata
        metadata.create_all(db)

        # from sqlalchemy.exc import IntegrityError
        # try:
        #     from strongr.schedulerdomain.model import Vm, VmState, Job, JobState
        #     vm = Vm()
        #     vm.cores = 4
        #     vm.ram = 32
        #     vm.vm_id = "7163a8cc-8d7d-409e-9360-52e6f6fe5b8a"
        #     vm.state = VmState.NEW
        #
        #     session = Gateways.sqlalchemy_session()
        #     session.add(vm)
        #     session.commit()
        # except IntegrityError:
        #     pass # continue when record already exists, we don't really care
        #
        # query_factory = SchedulerDomain.queryFactory()
        # query_bus = SchedulerDomain.schedulerService().getQueryBus()
        #
        # print(query_bus.handle(query_factory.newFindNodeWithAvailableResources(1, 1)))



class ScheduleJob:
    def __init__(self, image, script, job_id, scratch, cores, memory, secrets):
        self.image = image
        self.script = script
        self.job_id = job_id
        self.scratch = scratch
        self.cores = cores
        self.memory = memory
        self.secrets = secrets



from strongr.restdomain.query.oauth2 import RetrieveClient, RetrieveGrant,\
                                            RetrieveTokenByAccessToken, RetrieveTokenByRefreshToken

from strongr.core.exception import InvalidParameterException

class QueryFactory:
    """ This factory instantiates query objects to be sent to a rest querybus. """
    def newRetrieveClient(client_id):
        """ Generates a new RetrieveClient query

        :param client_id: the client id
        :type client_id: string

        :returns: A RetrieveClient query object
        :rtype: RetrieveClient
        """
        client_id = client_id.strip()
        if not len(client_id) > 0:
            raise InvalidParameterException('Client id is invalid'.format(cmd))

        return RetrieveClient(client_id=client_id)

    def newRetrieveGrant(self, client_id, code):
        """ Generates a new RetrieveGrant query

        :param client_id: the client id
        :type client_id: string

        :param code: the code
        :type code: string

        :returns: A RetrieveGrant query object
        :rtype: RetrieveGrant
        """
        client_id = client_id.strip()
        if not len(client_id) > 0:
            raise InvalidParameterException('Client id is invalid'.format(cmd))

        code = code.strip()
        if not len(code) > 0:
            raise InvalidParameterException('Code is invalid'.format(cmd))

        return RetrieveGrant(client_id=client_id, code=code)

    def newRetrieveTokenByAccessToken(self, access_token):
        """ Generates a new RetrieveTokenByAccessToken query

        :param access_token: the access token
        :type access_token: string

        :returns: A RetrieveTokenByAccessToken query object
        :rtype: RetrieveTokenByAccessToken
        """
        client_id = client_id.strip()
        if not len(client_id) > 0:
            raise InvalidParameterException('Client id is invalid'.format(cmd))

        return RetrieveTokenByAccessToken(access_token)

    def newRetrieveTokenByRefreshToken(self, refresh_token):
        """ Generates a new RetrieveTokenByRefreshToken query

        :param refresh_token: the refresh token
        :type refresh_token: string

        :returns: A RetrieveTokenByRefreshToken query object
        :rtype: RetrieveTokenByRefreshToken
        """
        client_id = client_id.strip()
        if not len(client_id) > 0:
            raise InvalidParameterException('Client id is invalid'.format(cmd))

        return RetrieveTokenByRefreshToken(refresh_token)



import os.path
import json

class JsonLoader():
    def getConfig(self, environment):
        configLocations = ['/etc/strongr/config.json', os.path.expanduser('~/.strongr/config.json'), 'config.json']
        output = {}

        envs = ['defaults', environment]

        for configLocation in configLocations:
            if os.path.isfile(configLocation):
                with open(configLocation) as jsonFile:
                    data = json.load(jsonFile)
                for env in envs:
                    if env in data:
                        for key in data[env]:
                            output[key] = data[env][key]
        return output



import strongr.core.gateways
import strongr.core
from strongr.schedulerdomain.model import Vm, VmState
from strongr.schedulerdomain.model.scalingdrivers import ScalingDriver


class VmReadyHandler(object):
    def __call__(self, command):
        session = strongr.core.gateways.Gateways.sqlalchemy_session()

        deadline = ScalingDriver.scaling_driver().get_vm_max_age(command.vm_id)

        session.query(Vm).filter(Vm.vm_id == command.vm_id).update({Vm.state: VmState.READY, Vm.deadline: deadline}, synchronize_session='fetch')



import strongr.core
import strongr.core.domain.schedulerdomain
import strongr.core.gateways

class RunEnqueuedJobsHandler:
    def __call__(self, command):
        # this command should be simplified at some point
        # finding jobs and a vm to run the job on could be
        # done in one sql query for example.

        SchedulerDomain = strongr.core.domain.schedulerdomain.SchedulerDomain

        schedulerService = SchedulerDomain.schedulerService()
        commandBus = schedulerService.getCommandBus()

        queryBus = schedulerService.getQueryBus()
        queryFactory = SchedulerDomain.queryFactory()

        commandFactory = SchedulerDomain.commandFactory()

        jobs = queryBus.handle(queryFactory.newRequestScheduledJobs()) # this query only gives us JobState enqueued and assigned jobs

        for job in jobs:
            # task is not running, let's try to execute it on an available node
            vm_id = queryBus.handle(queryFactory.newFindNodeWithAvailableResources(job.cores, job.ram))
            if vm_id == None:
                continue

            commandBus.handle(commandFactory.newStartJobOnVm(vm_id, job.job_id))



from strongr.core.middlewares.celery.envelope import Envelope

class CommandRouter:
    _routes = {}
    _envelopes = {}
    _worker_envelopes = []

    def append_route(self, name_list, commandbus):
        for name in name_list:
            self._routes[name] = commandbus

    def handle_command(self, command):
        name = command.__module__.split('.')[1] + '.' + command.__class__.__name__.lower()
        if name in self._routes:
            return self._routes[name].handle(command)

    def bind_all_routes(self, celery_app):
        for route in self._routes.keys():
            self.enable_route_for_command(celery_app, route)

    def enable_worker_route_for_command(self, celery_app, name):
        if name not in self._envelopes:
            self.enable_route_for_command(celery_app, name)
            self._worker_envelopes.append(name)

    def enable_route_for_command(self, celery_app, name):
        """
        Register command with celery
        """
        if name not in self._envelopes:
            # envelope self-registers with Celery
            celery_app.conf.task_routes = self._generate_celery_routes(name) # update routes in celery
            self._envelopes[name] = Envelope()
            if '.query.' in name.lower():
                self._envelopes[name].patch(celery_app, name, False)
            else:
                self._envelopes[name].patch(celery_app, name, True)

    def _generate_celery_routes(self, append):
        out = {}
        for route in self._routes.keys():
            out[route] = {'queue': route}

        if append not in out:
            out[append] = {'queue': append}

        return out

    def get_remotable_command_for(self, name):
        return self._envelopes[name]

    def has_remotable_command_registered(self, name):
        return name in self._envelopes and name not in self._worker_envelopes



from . import AbstractKeyValueStore

import time

class InMemoryStore(AbstractKeyValueStore):
    _store = {}
    _lastAccesed = {}

    def _logAccess(self, key):
        self._lastAccesed[key] = int(time.time())

    def _logRemove(self, key):
        self._lastAccesed.pop(key, None)

    def get(self, key):
        if key in self._store:
            self._logAccess(key)
            return self._store[key]
        return None

    def set(self, key, value):
        self._logAccess(key)
        self._store[key] = value

    def getAll(self):
        for key in self._store.keys():
            self._logAccess(key)
        return self._store.keys()

    def rem(self, key):
        self._logRemove(key)
        self._store.pop(key, None)

    def clear(self):
        self._logAccess = {}
        self._store = {}

    def append(self, key, value):
        if key in self._store:
            self._logAccess(key)
            self._store = self._store + value

    def exists(self, key):
        if key in self._store:
            return True
        return False



import dependency_injector.containers as containers
import dependency_injector.providers as providers

from strongr.schedulerdomain.service import SchedulerService
from strongr.schedulerdomain.factory import CommandFactory, QueryFactory

import strongr.core.domain.clouddomain

class SchedulerDomain(containers.DeclarativeContainer):
    """IoC container of service providers."""
    inter_domain_event_bindings = providers.Object({
        strongr.core.domain.clouddomain.CloudDomain.events()['jobfinished']: {  # command / query generators based on event
            'command': [
                (lambda event: SchedulerDomain.commandFactory().newJobFinished(event.job_id, event.ret, event.retcode))#,
                #(lambda event: SchedulerDomain.commandFactory().newRunEnqueuedJobs())
            ]
        },
        strongr.core.domain.clouddomain.CloudDomain.events()['vmnew']: {
            'command': [
                (lambda event: SchedulerDomain.commandFactory().newVmNew(event.vm_id, event.cores, event.ram))
            ]
        },
        strongr.core.domain.clouddomain.CloudDomain.events()['vmcreated']: {
            'command': [
                (lambda event: SchedulerDomain.commandFactory().newVmCreated(event.vm_id))
            ]
        },
        strongr.core.domain.clouddomain.CloudDomain.events()['vmready']: {
            'command': [
                (lambda event: SchedulerDomain.commandFactory().newVmReady(event.vm_id))
            ]
        },
        strongr.core.domain.clouddomain.CloudDomain.events()['vmdestroyed']: {
            'command': [
                (lambda event: SchedulerDomain.commandFactory().newVmDestroyed(event.vm_id))
            ]
        }
    })

    schedulerService = providers.Singleton(SchedulerService, inter_domain_event_bindings())
    commandFactory = providers.Singleton(CommandFactory)
    queryFactory = providers.Singleton(QueryFactory)



from salt.exceptions import SaltSystemExit

from strongr.clouddomain.handler.abstract.cloud import AbstractDestroyVmsHandler

import salt.cloud
import strongr.core

import logging

class DestroyVmsHandler(AbstractDestroyVmsHandler):
    def __call__(self, command):
        client = salt.cloud.CloudClient(strongr.core.Core.config().clouddomain.OpenNebula.salt_config + '/cloud')

        logger = logging.getLogger(self.__class__.__name__)

        ret = []

        # we are using _chunk_list generator because at some point we might want to remove more than 1 at a time
        for chunked_names in self._chunk_list(command.names, 1): # remove one by one, we are offloading this to amqp anyway
            try:
                ret.append(client.destroy(names=chunked_names))
            except SaltSystemExit as e:
                # An exception occured within salt. Normally vmdestroyed event would be published trough salt event system.
                # Assume VM is no longer there and broadcast vm destroyed event from here.
                # If it turns out the vm is still there but the error was triggered due to api rate limiting or flaky connection
                # the cleanup should remove the vm at a later time.
                inter_domain_event_factory = strongr.clouddomain.model.gateways.Gateways.inter_domain_event_factory()
                vmdestroyed_event = inter_domain_event_factory.newVmDestroyedEvent(chunked_names[0])
                strongr.core.Core.inter_domain_events_publisher().publish(vmdestroyed_event)
                logger.warning(e)

        return ret

    def _chunk_list(self, list, chunksize):
        for i in range(0, len(list), chunksize):
            yield list[i:i + chunksize]



from abc import ABCMeta, abstractmethod

class AbstractVmTemplateRetriever(object):
    __metaclass__ = ABCMeta

    @abstractmethod
    def get_vm_max_age(self, vm_name):
        pass

    @abstractmethod
    def get_templates(self):
        pass



from authlib.specs.rfc6749.grants import (
    ClientCredentialsGrant as _ClientCredentialsGrant
)

from strongr.core.gateways import Gateways
from strongr.restdomain.model.oauth2 import Token


class ClientCredentialsGrant(_ClientCredentialsGrant):
    TOKEN_ENDPOINT_AUTH_METHODS = [
        'client_secret_basic', 'client_secret_post'
    ]

    def create_access_token(self, token, client):
        item = Token(
            client_id=client.client_id,
            user_id=client.user_id,
            **token
        )
        session = Gateways.sqlalchemy_session()
        session.add(item)
        session.commit()



import statsd

from strongr.core.stats.abstractstats import AbstractStats

class StatsDDriver(AbstractStats):
    _statsd = None

    def __init__(self, config):
        self._statsd = statsd.StatsClient(**config)

    def decr(self, namespace, amount, rate=1):
        self._statsd.decr(namespace, amount, rate)

    def gauge(self, namespace, amount, rate=1, delta=False):
        self._statsd.gauge(namespace, amount, rate, delta)

    def set(self, namespace, arr):
        self._statsd.set(namespace, arr)

    def time(self, namespace):
        return self._statsd.timer(namespace)

    def timing(self, namespace, timems):
        self._statsd.timing(namespace, timems)

    def incr(self, namespace, amount, rate=1):
        self._statsd.incr(namespace, amount, rate)



from strongr.clouddomain.handler.abstract.cloud import AbstractDeployVmsHandler

import salt.cloud
import time

import strongr.core

import strongr.clouddomain.model.gateways

class DeployVmsHandler(AbstractDeployVmsHandler):
    def __call__(self, command):
        overrides = {}

        overrides['memory'] = command.ram * 1024
        overrides['cpu'] = command.cores
        overrides['vcpu'] = command.cores

        client = salt.cloud.CloudClient(strongr.core.Core.config().clouddomain.OpenNebula.salt_config + '/cloud')

        ret = []
        for chunked_names in self._chunk_list(command.names, 50):
            for name in chunked_names:
                vmnew_event = strongr.clouddomain.model.gateways.Gateways.inter_domain_event_factory().newVmNewEvent(name, command.cores, command.ram)
                strongr.core.Core.inter_domain_events_publisher().publish(vmnew_event)
            ret.append(client.profile(names=chunked_names, profile=command.profile, vm_overrides=overrides, parallel=True))
            #time.sleep(60) # api rate limiting removed from hpc cloud, testing with sleep turned off
            # this sleep is here because of HPCCloud API rate limiting
            # we should find a better solution at some point...

        return ret

    def _chunk_list(self, list, chunksize):
        for i in range(0, len(list), chunksize):
            yield list[i:i + chunksize]



import uuid

from sqlalchemy.orm import synonym

import strongr.core.gateways as gateways
from sqlalchemy import Column, String

from strongr.core.sqlalchemydatatypes.uuidtype import UUIDType
import strongr.restdomain.model.gateways

Base = gateways.Gateways.sqlalchemy_base()

class User(Base):
    __tablename__ = 'oauth_users'

    user_id = Column(String(64), primary_key=True, nullable=False, default=uuid.uuid4)
    username = Column(String(64), nullable=False)
    _password = Column('password', String(93)) # werkzeug generate_password_hash method='pbkdf2:sha256:80000' == 93 chars
    email = Column(String(250), nullable=False)

    @property
    def password(self):
        return self._password

    # update password-field using proper hash function
    @password.setter
    def password(self, value):
        self._password = strongr.restdomain.model.gateways.Gateways.generate_password_hash(password=value)

    def check_password(self, password):
        return strongr.restdomain.model.gateways.Gateways.check_password_hash(pwhash=self.password, password=password)


    # create a synonym so that _password and password are considered the same field by the mapper
    password = synonym('_password', descriptor=password)



from strongr.core.exception import InvalidParameterException

from strongr.secretsdomain.command import AddSecret, RemoveSecret

class CommandFactory:
    def new_add_secret(self, key, value):
        if len(key) == 0:
            raise InvalidParameterException('key len must be > 0')
        elif not isinstance(key, basestring):
            raise InvalidParameterException('key must be string type')

        if len(value) == 0:
            raise InvalidParameterException('value len must be > 0')
        elif not isinstance(value, basestring):
            raise InvalidParameterException('value must be string type')

        return AddSecret(key, value)

    def new_remove_secret(self, key):
        if len(key) == 0:
            raise InvalidParameterException('key len must be > 0')

        return RemoveSecret(key)



import strongr.core.domain.clouddomain
from sqlalchemy import and_, or_, func
from strongr.schedulerdomain.model import Vm, VmState, Job, JobState
from datetime import datetime, timedelta

import strongr.core
from strongr.schedulerdomain.model.scalingdrivers import ScalingDriver


class CleanupNodesHandler(object):
    def __call__(self, command):
        # VM's killed in the cloud are not yet synchronized to local db

        cloud_command_factory = strongr.core.domain.clouddomain.CloudDomain.commandFactory()
        cloud_command_bus = strongr.core.domain.clouddomain.CloudDomain.cloudService().getCommandBus()
        cloud_query_factory = strongr.core.domain.clouddomain.CloudDomain.queryFactory()
        cloud_query_bus = strongr.core.domain.clouddomain.CloudDomain.cloudService().getQueryBus()

        vm_templates = ScalingDriver.scaling_driver().get_templates()

        deadline = datetime.utcnow() + timedelta(minutes=-30) # give cloud domain time to provision a machine, if it isn't online by then it will probably never be
        session = strongr.core.gateways.Gateways.sqlalchemy_session()
        unprovisioned_vms_in_db = session.query(Vm).filter(and_(Vm.state.in_([VmState.NEW, VmState.PROVISION]), deadline > Vm.state_date)).all()
        vms_in_cloud = cloud_query_bus.handle(cloud_query_factory.newListDeployedVms())

        parallel_remove_list = []
        for vm in unprovisioned_vms_in_db:
            if vm.vm_id in vms_in_cloud['up'] or vm in vms_in_cloud['down']:
                vm.state = VmState.MARKED_FOR_DEATH
            else: # vm was never up or manually destroyed
                vm.state = VmState.FAILURE

        session.commit() # execute update queries

        vms_in_db = [vm[0] for vm in session.query(Vm.vm_id).filter(
            and_(Vm.state.in_([VmState.NEW, VmState.PROVISION, VmState.READY, VmState.MARKED_FOR_DEATH]))).all()]

        # cleanup unsynced / unregistered VM's
        for template in vm_templates:
            for vm in vms_in_cloud['up']:
                if vm not in vms_in_db and vm.startswith(template + '-'):
                    parallel_remove_list.append(vm)

            for vm in vms_in_cloud['down']:
                if vm not in vms_in_db and vm.startswith(template + '-'):
                    parallel_remove_list.append(vm)

        # check for VM's marked for death without jobs
        deadline = datetime.utcnow() + timedelta(seconds=-10) # vm's need to be marked for death for at least 10 secs before we clean them up else we run into race conditions
        subquery = session.query(Job.vm_id,
                                 func.count(Job.job_id).label('jobs'))\
                                .filter(
                                    Job.state.in_([JobState.ASSIGNED, JobState.RUNNING])
                                ).group_by(Job.vm_id).subquery('j')
        marked_for_death_vms = session.query(Vm)\
            .outerjoin(subquery, subquery.c.vm_id == Vm.vm_id)\
            .filter(and_(Vm.state == VmState.MARKED_FOR_DEATH, or_(subquery.c.jobs == None, subquery.c.jobs == 0),  deadline > Vm.state_date))\
            .all()

        for vm in marked_for_death_vms:
            parallel_remove_list.append(vm.vm_id)

        # check for expired VM's
        timenow = datetime.utcnow()
        expired_vms = session.query(Vm).filter(and_(timenow > Vm.deadline, Vm.state.in_([VmState.NEW, VmState.PROVISION, VmState.READY]))).all()
        for vm in expired_vms:
            vm.state = VmState.MARKED_FOR_DEATH
        session.commit()

        # send vm destroy command to clouddomain
        if len(parallel_remove_list) > 0:
            command = cloud_command_factory.newDestroyVms(parallel_remove_list)
            cloud_command_bus.handle(command)



import unittest

import strongr.clouddomain.model.gateways

import strongr.core
import strongr.core.domain.clouddomain

class TestInterDomainEvent(unittest.TestCase):
    def test_salt_job_finished_escalation_to_inter(self):
        intra_domain_event_factory = strongr.clouddomain.model.gateways.Gateways.intra_domain_event_factory()
        cloudService = strongr.core.domain.clouddomain.CloudDomain.cloudService() # this initializes event subscribers within the domain

        event = intra_domain_event_factory.newSaltJobFinished('1', 'Test return', 0)

        strongr.core.Core.inter_domain_events_publisher().subscribe(strongr.core.domain.clouddomain.CloudDomain.events()['jobfinished'], (
            lambda event: self.assertTrue(True)
        ))

        strongr.clouddomain.model.gateways.Gateways.intra_domain_events_publisher().publish(event)



from strongr.core.domain.schedulerdomain import SchedulerDomain
from strongr.core.domain.clouddomain import CloudDomain
from .wrapper import Command

import time
import schedule
from celery import Celery
import strongr.core

class RunResourceManager(Command):
    """
    Runs the resource manager. This should be done on your salt master.

    resourcemanager:run
    """
    def handle(self):
        schedulerService = SchedulerDomain.schedulerService()
        commandFactory = SchedulerDomain.commandFactory()

        command_bus = schedulerService.getCommandBus()
        run_enqueued_jobs_command = commandFactory.newRunEnqueuedJobs()
        check_scaling_command = commandFactory.newCheckScaling()
        cleanup_nodes = commandFactory.newCleanupNodes()
        cleanup_jobs = commandFactory.newCleanupOldJobs()
        log_stats = commandFactory.newLogStats()

        CloudDomain.cloudService() # init clouddomain

        self.info('Running.')

        if hasattr(CloudDomain.cloudService(), 'start_reactor'): # ugly way for instantiating salt reactor
            CloudDomain.cloudService().start_reactor()  # salt reactor really shouldn't be initialised here
            self.info('Salt reactor started!')
        else:
            self.warning('No reactor, this is ok if you are running the mockcloud.')

        if hasattr(strongr.core.Core.config(), 'celery'): # don't load celery if it's not configured
            celery = Celery('strongr', broker=strongr.core.Core.config().celery.broker, backend=strongr.core.Core.config().celery.backend)

            remotable_commands = strongr.core.Core.config().celery.remotable_commands.as_dict()
            for domain in remotable_commands:
                for command in remotable_commands[domain]:
                    strongr.core.Core.command_router().enable_route_for_command(celery, '{}.{}'.format(domain, command))
        else:
            self.warning('Celery not configured, skipping')

        schedule.every(1).seconds.do(command_bus.handle, run_enqueued_jobs_command)
        schedule.every(5).seconds.do(command_bus.handle, check_scaling_command)
        schedule.every(10).seconds.do(command_bus.handle, log_stats)
        schedule.every(1).minutes.do(command_bus.handle, cleanup_nodes)
        schedule.every(30).minutes.do(command_bus.handle, cleanup_jobs)

        while True:
            schedule.run_pending()
            time.sleep(.5)



from strongr.secretsdomain.model import Secret
import strongr.core.gateways

import uuid

class AddSecretHandler():
    def __call__(self, command):
        session = strongr.core.gateways.Gateways.sqlalchemy_session()

        secret = session.query(Secret).filter(Secret.key == command.key).first()

        if secret is None:
            secret = Secret()

        secret.key = command.key
        secret.value = command.value

        session.add(secret)



import strongr.core.gateways as gateways
from sqlalchemy import Column, ForeignKey, Integer, String, Enum, DateTime, func
from sqlalchemy.orm import relationship, synonym

from strongr.schedulerdomain.model.vmstate import VmState

Base = gateways.Gateways.sqlalchemy_base()

class Vm(Base):
    __tablename__ = 'vms'

    vm_id = Column(String(64), primary_key=True)
    cores = Column(Integer)
    ram = Column(Integer)
    deadline = Column(DateTime())

    jobs = relationship('Job', back_populates='vm')

    _state = Column('state', Enum(VmState))

    # In classical SQL we would put a trigger to update this field with NOW() if the state-field is updated.
    # SQLAlchemy has no way to write triggers without writing platform-dependent SQL at the time of writing.
    # Instead we use a setter on the state-field, this setter updates the state_date as well.
    # The disadvantage of this solution is that other apps need to implement logic like this as well making
    # the solution less portable.
    state_date = Column(DateTime())

    @property
    def state(self):
        return self._state

    # update state_date-field as well when we update state-field
    @state.setter
    def state(self, value):
        self._state = value
        self.state_date = func.now()

    # create a synonym so that _state and state are considered the same field by the mapper
    state = synonym('_state', descriptor=state)



from strongr.core.exception import InvalidParameterException

from strongr.clouddomain.event.inter.jobfinished import JobFinished
from strongr.clouddomain.event.inter.vmdestroyed import VmDestroyed
from strongr.clouddomain.event.inter.vmready import VmReady
from strongr.clouddomain.event.inter.vmcreated import VmCreated
from strongr.clouddomain.event.inter.vmnew import VmNew

# try-except for py 2 / 3 compatibility
try:
    basestring
except NameError:
    basestring = str

class InterDomainEventFactory(object):
    def newJobFinishedEvent(self, job_id, ret, retcode):
        if not isinstance(job_id, basestring) or len(job_id.strip()) == 0:
            raise InvalidParameterException('job_id is invalid')
        elif not isinstance(ret, basestring) or len(ret.strip()) == 0:
            raise InvalidParameterException('ret is invalid')
        elif not isinstance(retcode, int):
            raise InvalidParameterException('retcode is invalid')

        return JobFinished(job_id, ret, retcode)

    def newVmDestroyedEvent(self, job_id):
        if not isinstance(job_id, basestring) or len(job_id.strip()) == 0:
            raise InvalidParameterException('job_id not valid')

        return VmDestroyed(job_id)

    def newVmReadyEvent(self, job_id):
        if not isinstance(job_id, basestring) or len(job_id.strip()) == 0:
            raise InvalidParameterException('job_id not valid')

        return VmReady(job_id)

    def newVmCreatedEvent(self, job_id):
        if not isinstance(job_id, basestring) or len(job_id.strip()) == 0:
            raise InvalidParameterException('job_id not valid')

        return VmCreated(job_id)

    def newVmNewEvent(self, job_id, cores, ram):
        if not isinstance(job_id, basestring) or len(job_id.strip()) == 0:
            raise InvalidParameterException('job_id not valid')
        elif not isinstance(cores, int) or cores <= 0:
            raise InvalidParameterException('cores not valid')
        elif not isinstance(ram, int) or cores <= 0:
            raise InvalidParameterException('ram not valid')

        return VmNew(job_id, cores, ram)



from strongr.core.abstracts.abstractservice import AbstractService

from strongr.secretsdomain.command import AddSecret, RemoveSecret
from strongr.secretsdomain.handler import AddSecretHandler, RemoveSecretHandler

from strongr.secretsdomain.query import GetSecret, ListSecrets
from strongr.secretsdomain.handler import GetSecretHandler, ListSecretsHandler

class SecretService(AbstractService):
    _command_bus = None
    _query_bus = None

    def register_models(self):
        from strongr.secretsdomain.model import Secret
        # importing alone is enough for registration

    def get_command_bus(self):
        if self._command_bus is None:
            self._command_bus = self._make_default_commandbus({
                        AddSecretHandler: AddSecret,
                        RemoveSecretHandler: RemoveSecret
                    })
        return self._command_bus

    def get_query_bus(self):
        if self._query_bus is None:
            self._query_bus = self._make_default_querybus({
                    GetSecretHandler: GetSecret,
                    ListSecretsHandler: ListSecrets
                })
        return self._query_bus



from strongr.core.domain.schedulerdomain import SchedulerDomain
from .wrapper import Command

import json

class GetFinishedJobsCommand(Command):
    """
    Returns all finished jobs.

    jobs:finished
    """
    def handle(self):
        schedulerService = SchedulerDomain.schedulerService()
        queryFactory = SchedulerDomain.queryFactory()

        query = queryFactory.newRequestFinishedJobs()
        result = schedulerService.getQueryBus().handle(query)
        print(json.dumps(result))



import threading

import fnmatch
import salt.config
import salt.utils.event

import strongr.core

import strongr.clouddomain.factory.intradomaineventfactory
import strongr.clouddomain.factory.interdomaineventfactory
import strongr.clouddomain.model.gateways

import logging

class SaltEventTranslator(threading.Thread):
    def run(self):
        opts = salt.config.client_config(strongr.core.Core.config().clouddomain.OpenNebula.salt_config + '/master')
        inter_domain_event_factory = strongr.clouddomain.model.gateways.Gateways.inter_domain_event_factory()

        event = salt.utils.event.get_event(
            'master',
            sock_dir=opts['sock_dir'],
            transport=opts['transport'],
            opts=opts)

        while True:
            ret = event.get_event(full=True)
            if ret is None:
                continue

            try:
                if fnmatch.fnmatch(ret['tag'], 'salt/job/*/ret/*'):
                    data = ret['data']
                    if 'jid' in data and 'return' in data and 'retcode' in data:
                        job_finished_event = inter_domain_event_factory.newJobFinishedEvent(data['jid'], data['return'], data['retcode'])
                        strongr.core.Core.inter_domain_events_publisher().publish(job_finished_event)
                elif fnmatch.fnmatch(ret['tag'], 'salt/cloud/*/creating'):
                    data = ret['data']
                    if 'name' in data:
                        vmcreated_event = inter_domain_event_factory.newVmCreatedEvent(data['name'])
                        strongr.core.Core.inter_domain_events_publisher().publish(vmcreated_event)
                elif fnmatch.fnmatch(ret['tag'], 'salt/cloud/*/created'):
                    data = ret['data']
                    if 'name' in data:
                        vmready_event = inter_domain_event_factory.newVmReadyEvent(data['name'])
                        strongr.core.Core.inter_domain_events_publisher().publish(vmready_event)
                elif fnmatch.fnmatch(ret['tag'], 'salt/cloud/*/destroyed'):
                    data = ret['data']
                    if 'name' in data:
                        vmdestroyed_event = inter_domain_event_factory.newVmDestroyedEvent(data['name'])
                        strongr.core.Core.inter_domain_events_publisher().publish(vmdestroyed_event)
            except Exception as e: # thread must always continue running
                logging.getLogger('SaltEventTranslator').warning(str(e))
                pass



import dependency_injector.containers as containers
import dependency_injector.providers as providers

from strongr.clouddomain.event.inter.jobfinished import JobFinished
from strongr.clouddomain.event.inter.vmcreated import VmCreated
from strongr.clouddomain.event.inter.vmready import VmReady
from strongr.clouddomain.event.inter.vmdestroyed import VmDestroyed
from strongr.clouddomain.event.inter.vmnew import VmNew

from strongr.clouddomain.factory import CommandFactory, QueryFactory
from strongr.clouddomain.service import CloudServices

import strongr.core

class CloudDomain(containers.DeclarativeContainer):
    """IoC container of service providers."""
    cloudService = providers.Singleton(CloudServices().getCloudServiceByName, strongr.core.Core.config().clouddomain.driver)
    commandFactory = providers.Singleton(CommandFactory)
    queryFactory = providers.Singleton(QueryFactory)

    events = providers.Object({
        'jobfinished': JobFinished,
        'vmnew': VmNew,
        'vmcreated': VmCreated,
        'vmready': VmReady,
        'vmdestroyed': VmDestroyed
    })



from strongr.clouddomain.handler.abstract.cloud import AbstractRunJobHandler
import strongr.core

import salt.client

class RunJobHandler(AbstractRunJobHandler):
    def __call__(self, command):
        local = salt.client.LocalClient()
        local.cmd_async(command.host, 'cmd.run', ['cmd={}'.format(command.script), 'runas={}'.format(strongr.core.Core.config().clouddomain.OpenNebula.runas)], jid=command.job_id)



import unittest

from strongr.schedulerdomain.model import Job, JobState
import strongr.core.gateways
import strongr.core.domain.schedulerdomain

class TestJobFinished(unittest.TestCase):
    def test_job_finished(self):
        session = strongr.core.gateways.Gateways.sqlalchemy_session()

        session.query(Job).filter(Job.job_id == 'unittestjob').delete()

        job = Job()
        job.job_id = 'unittestjob'
        job.cmd = 'ls -l'
        job.cores = 1
        job.ram = 1
        job.state = JobState.ENQUEUED

        session = strongr.core.gateways.Gateways.sqlalchemy_session()
        session.add(job)
        session.commit()

        factory = strongr.core.domain.schedulerdomain.SchedulerDomain.commandFactory()
        job_finished = factory.newJobFinished(job.job_id, 'Testing jobfinished command', 0)
        strongr.core.domain.schedulerdomain.SchedulerDomain.schedulerService().getCommandBus().handle(job_finished)

        self.assertEqual(job.state, JobState.FINISHED)

        session.query(Job).filter(Job.job_id == 'unittestjob').delete()



from cmndr import Middleware
from sqlalchemy.exc import SQLAlchemyError

import logging


class TransactionMiddleware(Middleware):
    def execute(self, command, next_callable):
        import strongr.core
        if hasattr(strongr.core.Core.config(), 'db') and hasattr(strongr.core.Core.config().db, 'engine'):
            import strongr.core.gateways

            session = strongr.core.gateways.Gateways.sqlalchemy_session()

            try:
                ret = next_callable(command)
                session.commit()
                return ret
            except SQLAlchemyError as e:
                session.rollback()
                logging.getLogger("Transaction Middleware").warning(e)
                #raise e
        else:
            # db not configured yet, ignore transactions for now
            return next_callable(command)



from strongr.restdomain.command.oauth2 import AppendGrant

from strongr.core.exception import InvalidParameterException

class CommandFactory:
    """ This factory instantiates command objects to be sent to a rest commandbus. """
    def newAppendGrant(client_id, code, redirect_uri, scope, user_id, expires):
        """ Generates a new AppendGrant command

        :param client_id: the client id
        :type client_id: string

        :param code: the code
        :type code: string

        :param redirect_uri: the redirect uri
        :type redirect_uri: string

        :param scope: the scope
        :type scope: string

        :param user_id: the user_id
        :type user_id: string

        :param expires: when should the grant expire?
        :type expires: int

        :returns: An AppendGrant command object
        :rtype: AppendGrant
        """
        client_id = client_id.strip()
        if not len(client_id) > 0:
            raise InvalidParameterException('Client id is invalid'.format(cmd))

        code = code.strip()
        if not len(code) > 0:
            raise InvalidParameterException('Code is invalid'.format(cmd))

        redirect_uri = redirect_uri.strip()
        if not len(redirect_uri) > 0:
            raise InvalidParameterException('Redirect uri is invalid'.format(cmd))

        scope = scope.strip()
        if not len(scope) > 0:
            raise InvalidParameterException('Scope is invalid'.format(cmd))

        user_id = user_id.strip()
        if not len(user_id) > 0:
            raise InvalidParameterException('User id is invalid'.format(cmd))


        if not expires >= 0:
            raise InvalidParameterException('Expires is invalid'.format(cmd))

        return AppendGrant(client_id, code, redirect_uri, scope, user_id, expires)



from strongr.core.domain.secretdomain import SecretDomain
from .wrapper import Command


class GetSecret(Command):
    """
    Returns a jobs stdout

    secret:get
        {key : The key of the secret}
    """
    def handle(self):
        service = SecretDomain.secret_service()
        query_bus = service.get_query_bus()
        query_factory = SecretDomain.query_factory()

        result = query_bus.handle(query_factory.new_get_secret(self.argument('key')))

        print(result)



from authlib.specs.rfc6749.grants import (
    RefreshTokenGrant as _RefreshTokenGrant
)

from strongr.core.gateways import Gateways
from strongr.restdomain.model.oauth2 import Token


class RefreshTokenGrant(_RefreshTokenGrant):
    def authenticate_token(self, refresh_token):
        item = Token.query.filter_by(refresh_token=refresh_token).first()
        # define is_refresh_token_expired by yourself
        if item and not item.is_refresh_token_expired():
            return item

    def create_access_token(self, token, authenticated_token):
        # issue a new token to replace the old one, you can also update
        # the ``authenticated_token`` instead of issuing a new one
        item = Token(
            client_id=authenticated_token.client_id,
            user_id=authenticated_token.user_id,
            **token
        )
        session = Gateways.sqlalchemy_session()
        session.add(item)
        session.delete(authenticated_token)
        session.commit()



from abc import ABCMeta, abstractmethod

from cmndr import CommandBus
from cmndr.handlers import CommandHandler
from cmndr.handlers.inflectors import CallableInflector
from cmndr.handlers.locators import LazyLoadingInMemoryLocator
from cmndr.handlers.nameextractors import ClassNameExtractor

import strongr.core

from strongr.core.middlewares.celery.celerymiddleware import CeleryMiddleware
from strongr.core.middlewares.logging.loggingmiddleware import LoggingMiddleware
from strongr.core.middlewares.stats.statsmiddleware import StatsMiddleware
from strongr.core.middlewares.transaction.transactionmiddleware import TransactionMiddleware

class AbstractService():
    __metaclass__ = ABCMeta

    def __init__(self, inter_domain_event_bindings=None):
        if inter_domain_event_bindings is not None:
            for event in inter_domain_event_bindings:
                if 'command' in inter_domain_event_bindings[event]:
                    for command_generator in inter_domain_event_bindings[event]['command']:
                        strongr.core.Core.inter_domain_events_publisher().subscribe(event, (
                        lambda event, command_generator=command_generator: self.getCommandBus().handle(command_generator(event))))
                if 'query' in inter_domain_event_bindings[event]:
                    for query_generator in inter_domain_event_bindings[event]['query']:
                        strongr.core.Core.inter_domain_events_publisher().subscribe(event, (
                        lambda event, command_generator=command_generator: self.getQueryBus().handle(query_generator(event))))

    @abstractmethod
    def register_models(self):
        """
        Every service should register its models with sqlalchemy
        :return:
        """
        pass

    def _default_middlewares(self, will_return_values, enable_stats=True):
        middlewares = [
            TransactionMiddleware(),
            LoggingMiddleware(),
            CeleryMiddleware(will_return_values)
        ]

        if enable_stats:
            import strongr.core.gateways
            middlewares.append(StatsMiddleware(strongr.core.gateways.Gateways.stats()))

        return middlewares

    def _make_default_querybus(self, mappings, middlewares=None, enable_stats=True):
        return self._make_default_bus(mappings, middlewares, True, enable_stats)

    def _make_default_commandbus(self, mappings, middlewares=None, enable_stats=True):
        return self._make_default_bus(mappings, middlewares, False, enable_stats)

    def _make_default_bus(self, mappings, middlewares, will_return_values, enable_stats=True):
        handlers = {}
        remotable_mappings = []
        for key in mappings.keys():
            handlers[key] = mappings[key].__name__
            remotable_mappings.append(mappings[key].__module__.split('.')[1] + '.' + mappings[key].__name__.lower())

        extractor = ClassNameExtractor()
        locator = LazyLoadingInMemoryLocator(handlers)
        inflector = CallableInflector()
        handler = CommandHandler(extractor, locator, inflector)
        if middlewares is not None:
            bus = CommandBus(middlewares + self._default_middlewares(will_return_values, enable_stats) + [handler])
        else:
            bus = CommandBus(self._default_middlewares(will_return_values, enable_stats) + [handler])

        # this is needed to get remotable (celery) commands working
        # it takes care of routing celery tasks to the appropriate command bus
        strongr.core.Core.command_router().append_route(remotable_mappings, bus)

        return bus



import time
import strongr.core.gateways
import logging

class RedisLock(object):
    def __init__(self, name, namespace, timeout):
        self._timeout = timeout
        self._name = name
        self._redis_namespace = namespace
        self._redis = strongr.core.gateways.Gateways.redis()
        self._logger = logging.getLogger('redislock')

    def exists(self):
        return self._redis.exists(self._redis_namespace + self._name)

    def __enter__(self):
        timeout_after = int(time.time()) + self._timeout
        self._logger.debug('Attempting to acquire lock {}'.format(self._name))
        while not self._redis.setnx(self._redis_namespace + self._name, True) and int(time.time()) > timeout_after:
            time.sleep(.1)
        else:
            self._logger.debug('Lock acquired on {}'.format(self._name))
            return self

        self._logger.debug('Failed to acquire lock on {}'.format(self._name))
        raise IOError('Could not aquire lock for {}'.format(self._name))

    def __exit__(self, type, value, traceback):
        self._logger.debug('Releasing lock on {}'.format(self._name))
        self._redis.delete(self._redis_namespace + self._name)



from strongr.schedulerdomain.model.scalingdrivers.abstract import AbstractScaleOut, AbstractScaleIn, \
    AbstractVmTemplateRetriever

import uuid

import strongr.core
import strongr.core.gateways
import logging
import strongr.core.domain.schedulerdomain

from sqlalchemy import func, and_, or_

from strongr.schedulerdomain.model import JobState, Job, Vm, VmState

from datetime import datetime, timedelta

# The most basic scaler.

class SimpleScaler(AbstractScaleIn, AbstractScaleOut, AbstractVmTemplateRetriever):
    def __init__(self, config, *args, **kwargs):
        super(SimpleScaler, self).__init__(*args, **kwargs)
        self._config = config

    def get_vm_max_age(self, vm_name):
        template = vm_name.vm_id.split('-')[0]
        if len(template) > 0:
            templates = self.get_templates()
            for t in templates:
                if t == template:
                    deadline = datetime.utcnow() + timedelta(seconds=(templates[t]['maxage'] if hasattr(templates[t], 'maxage') else 900))
        else:
            deadline = datetime.utcnow() + timedelta(minutes=15)

        return deadline

    def get_templates(self):
        return self._config.schedulerdomain.simplescaler.templates.as_dict()

    def scalein(self, command):
        if strongr.core.gateways.Gateways.lock('scaleout-lock').exists():
            return # only every run one of these commands at once

        with strongr.core.gateways.Gateways.lock('scaleout-lock'):  # only ever run one of these commands at once
            logger = logging.getLogger('schedulerdomain.' + self.__class__.__name__)

            session = strongr.core.gateways.Gateways.sqlalchemy_session()

            # subquery to see whats already running on vm
            subquery1 = session.query(Job.vm_id, func.count(Job.job_id).label('jobs'), func.sum(Job.cores).label('cores'), func.sum(Job.ram).label('ram')).filter(
                Job.state.in_([JobState.RUNNING])).group_by(Job.vm_id).subquery('j')

            subquery2 = session.query(Job.vm_id, func.max(Job.state_date).label('last_job_date')).filter(Job.state.in_([JobState.FAILED, JobState.FINISHED, JobState.RUNNING])).group_by(Job.vm_id).subquery('i')

            results = session.query(Vm.vm_id.label('vm_id'), subquery1.c.jobs.label('job_count'), subquery2.c.last_job_date) \
                .outerjoin(subquery1, subquery1.c.vm_id == Vm.vm_id) \
                .outerjoin(subquery2, subquery2.c.vm_id == Vm.vm_id) \
                .filter(
                and_(
                    or_(
                        and_(  # case 1 - vm with jobs, check if vm has about half capacity available
                            Vm.cores - subquery1.c.cores >= Vm.cores / 2,
                            Vm.ram - subquery1.c.ram >= Vm.ram / 2
                        ),
                        and_(  # case 2 - vm with no jobs
                            subquery1.c.cores == None,
                            subquery1.c.ram == None,
                        )
                    ),
                    Vm.state.in_([VmState.READY])  # vm should be in state ready
                )
            ).all()

            if not results:
                return # no VM's to scalein

            deadline = datetime.utcnow() + timedelta(minutes=-10)

            vms_to_update = []
            mark_for_death_counter = 0
            for vm in results:
                if (vm[1] is None or vm[1] == 0) and (vm[2] is None or deadline > vm[2]):
                    vms_to_update.append(vm[0])
                elif deadline > vm[2]:
                    if mark_for_death_counter % 2 == 0:
                        vms_to_update.append(vm)
                    mark_for_death_counter += 1

            if len(vms_to_update) > 0:
                    session.commit()
                    session.query(Vm).filter(Vm.vm_id.in_(vms_to_update)).update({Vm.state: VmState.MARKED_FOR_DEATH}, synchronize_session='fetch')

    def scaleout(self, command):
        cores = command.cores
        ram = command.ram

        if strongr.core.gateways.Gateways.lock('scaleout-lock').exists():
            return # only every run one of these commands at once

        with strongr.core.gateways.Gateways.lock('scaleout-lock'):  # only ever run one of these commands at once
            config = self._config
            logger = logging.getLogger('schedulerdomain.' + self.__class__.__name__)

            query_factory = strongr.core.domain.schedulerdomain.SchedulerDomain.queryFactory()
            query_bus = strongr.core.domain.schedulerdomain.SchedulerDomain.schedulerService().getQueryBus()

            templates = dict(config.schedulerdomain.simplescaler.templates.as_dict()) # make a copy because we want to manipulate the list

            active_vms = query_bus.handle(query_factory.newRequestVms([VmState.NEW, VmState.PROVISION, VmState.READY]))

            provision_counter = 0
            for vm in active_vms:
                if vm.state in [VmState.NEW, VmState.PROVISION]:
                    cores -= vm.cores
                    ram -= vm.ram
                    provision_counter += 1
                # template = vm.vm_id.split('-')[0]
                # if template in templates:
                #     if 'spawned' in templates[template]:
                #         templates[template]['spawned'] += 1
                #     else:
                #         templates[template]['spawned'] = 1

            if provision_counter >= 6:
                # don't provision more than 6 VM's at the same time
                return

            # for template in list(templates): # make copy of list so that we can edit original
            #     if 'spawned' in templates[template] and templates[template]['spawned'] >= templates[template]['spawned-max']:
            #         del(templates[template]) # we already have the max amount of vms for this template

            # if not templates:
            #     return # max env size reached or no templates defined in config

            if cores <= 0 or cores < config.schedulerdomain.simplescaler.scaleoutmincoresneeded:
                return

            if ram <= 0 or ram < config.schedulerdomain.simplescaler.scaleoutminramneeded:
                return

            for template in templates:
                templates[template]['distance'] = templates[template]['ram'] / templates[template]['cores']

            ram_per_core_needed = ram / cores

            # find best fit based on templates

            # first we calculate the distance based on optimal mem / core distribution
            distances = {}
            for template in templates:
                distance = abs(templates[template]['distance'] - ram_per_core_needed)
                if distance not in distances:
                    distances[distance] = []
                distances[distance].append(template)

            min_distance_templates = distances[min(distances)] # get templates with least distance

            # now find the templates with resources that best fit what we need
            # we simply select the best fitting template with the most resources
            template = min(min_distance_templates, key=lambda key: (cores - templates[key]['cores']) + (ram - templates[key]['ram']))

            #template = min(templates, key=lambda key: abs(templates[key]['distance'] - ram_per_core_needed))

            # scaleout by one instance
            cloudService = strongr.core.domain.clouddomain.CloudDomain.cloudService()
            cloudCommandFactory = strongr.core.domain.clouddomain.CloudDomain.commandFactory()
            cloudProviderName = config.clouddomain.driver
            profile = getattr(config.clouddomain, cloudProviderName).default_profile if 'profile' not in templates[template] else templates[template]['profile']
            deployVmsCommand = cloudCommandFactory.newDeployVms(names=[template + '-' + str(uuid.uuid4())], profile=profile, cores=templates[template]['cores'], ram=templates[template]['ram'])

            cloudCommandBus = cloudService.getCommandBus()

            logger.info('Deploying VM {0} cores={1} ram={2}GiB'.format(deployVmsCommand.names[0], templates[template]['cores'], templates[template]['ram']))

            cloudCommandBus.handle(deployVmsCommand)



from strongr.core import Core
from strongr.core.domain.clouddomain import CloudDomain
from .wrapper import Command

import uuid

class DeploySingleCommand(Command):
    """
    Deploys a VM in the cloud. A first step towards elasticity.

    deploy:single
    """
    def handle(self):
        cloudService = CloudDomain.cloudService()
        commandFactory = CloudDomain.commandFactory()

        cores = int(self.ask('How many processing cores should the VM have? (default 1): ', 1))
        ram = int(self.ask('How much memory in GiB should the VM have? (default 4): ', 4))
        name = self.ask('What is the name of the VM? (default generated): ', 'worker-' + str(uuid.uuid4()))

        if not (cores > 0 and ram > 0 and len(name) > 0):
            # TODO: put something sensible in here, this is just a placeholder
            self.error('Invalid input')
            return

        scaling_driver = Core.config().schedulerdomain.scalingdriver
        driver_config = getattr(Core.config().schedulerdomain, scaling_driver)
        profile = driver_config['default_profile']
        deployVmsCommand = commandFactory.newDeployVms(names=[name], profile=profile, cores=cores, ram = ram)

        commandBus = cloudService.getCommandBus()

        self.info('Deploying VM {0} cores={1} ram={2}GiB'.format(name, cores, ram))

        commandBus.handle(deployVmsCommand)



from strongr.clouddomain.handler.abstract.cloud import AbstractDeployVmsHandler

import salt.cloud
import time

import strongr.core

import strongr.clouddomain.model.gateways

class DeployVmsHandler(AbstractDeployVmsHandler):
    def __call__(self, command):
        overrides = {}

        overrides['memory'] = command.ram * 1024
        overrides['cpu'] = command.cores
        overrides['vcpu'] = command.cores

        client = salt.cloud.CloudClient(strongr.core.Core.config().clouddomain.Azure.salt_config + '/cloud')

        ret = []
        for name in command.names:
            vmnew_event = strongr.clouddomain.model.gateways.Gateways.inter_domain_event_factory().newVmNewEvent(name, command.cores, command.ram)
            strongr.core.Core.inter_domain_events_publisher().publish(vmnew_event)
        ret.append(client.profile(names=command.names, profile=command.profile, vm_overrides=overrides, parallel=True))

        return ret



from cmndr import Middleware

import logging

class LoggingMiddleware(Middleware):

    def execute(self, command, next_callable):
        cmd_dict = command.__dict__

        splitted_module = command.__module__.split('.')
        domain = splitted_module[1]
        log_string = '{} {}'.format(' '.join(splitted_module[-2:]), cmd_dict if cmd_dict else '')
        logging.getLogger(domain).info(log_string)
        ret = next_callable(command)
        return ret



from strongr.schedulerdomain.model.scalingdrivers import ScalingDriver

class ScaleInHandler(object):
    def __call__(self, command):
        # forward to configured scalein driver
        ScalingDriver.scaling_driver().scalein(command)




import json

from strongr.core.domain.schedulerdomain import SchedulerDomain
from .wrapper import Command

class RequestScheduledTasks(Command):
    """
    Shows the task queue.

    job:list
    """
    def handle(self):
        schedulerService = SchedulerDomain.schedulerService()
        queryFactory = SchedulerDomain.schedulerQueryFactory()

        query = queryFactory.newRequestScheduledJobs()
        result = schedulerService.getQueryBus().handle(query)
        print(json.dumps(result))




import strongr.core

import strongr.core.domain.clouddomain
import strongr.core.gateways

from strongr.schedulerdomain.model import Job, Vm, VmState, JobState

from sqlalchemy.sql import func
from sqlalchemy import and_, or_

class FindNodeWithAvailableResourcesHandler:
    _timeout = 0

    def __call__(self, query):
        db = strongr.core.gateways.Gateways.sqlalchemy_session()

        # subquery to see whats already running on vm
        subquery = db.query(Job.vm_id, func.sum(Job.cores).label('cores'), func.sum(Job.ram).label('ram')).filter(Job.state.in_([JobState.RUNNING])).group_by(Job.vm_id).subquery('j')

        query = db.query(Vm.vm_id)\
            .outerjoin(subquery, subquery.c.vm_id == Vm.vm_id)\
            .filter(
                and_(
                    or_(
                        and_( # case 1 - vm with jobs, check if vm has enough availabale resources to run the job
                            Vm.cores - subquery.c.cores >= query.cores,
                            Vm.ram - subquery.c.ram >= query.ram
                        ),
                        and_( # case 2 - vm with no jobs, check if vm has enough available resources to run the job
                            subquery.c.cores == None,
                            subquery.c.ram == None,
                            Vm.cores >= query.cores,
                            Vm.ram >= query.ram
                        )
                    ),
                    Vm.state.in_([VmState.READY]) # vm should be in state ready before we push jobs to it
                )
            ).order_by(Vm.ram - subquery.c.ram)

        results = query.all()

        if len(results) == 0:
            return None
        return results[0].vm_id



from abc import ABCMeta, abstractmethod

class AbstractKeyValueStore:
    __metaclass__ = ABCMeta

    @abstractmethod
    def get(self, key):
        pass

    @abstractmethod
    def set(self, key, value):
        pass

    @abstractmethod
    def getAll(self):
        pass

    @abstractmethod
    def rem(self, key):
        pass

    @abstractmethod
    def clear(self):
        pass

    @abstractmethod
    def append(self, key, value):
        pass

    @abstractmethod
    def exists(self, key):
        pass



from strongr.core.abstracts.abstractservice import AbstractService

from strongr.restdomain.command.oauth2 import AppendGrant
from strongr.restdomain.handler.oauth2 import AppendGrantHandler

from strongr.restdomain.query.oauth2 import RetrieveClient, RetrieveGrant,\
                                            RetrieveTokenByAccessToken, RetrieveTokenByRefreshToken
from strongr.restdomain.handler.oauth2 import RetrieveClientHandler, RetrieveGrantHandler,\
                                        RetrieveTokenByAccessTokenHandler, RetrieveTokenByRefreshTokenHandler


class Oauth2Service(AbstractService):
    _command_bus = None
    _query_bus = None

    def register_models(self):
        import strongr.restdomain.model.oauth2
        # importing alone is enough for registration

    def getCommandBus(self):
        if self._command_bus is None:
            self._command_bus = self._make_default_commandbus({
                    AppendGrantHandler: AppendGrant,
                    })
        return self._command_bus

    def getQueryBus(self):
        if self._query_bus is None:
            self._query_bus = self._make_default_querybus({
                    RetrieveClientHandler: RetrieveClient,
                    RetrieveGrantHandler: RetrieveGrant,
                    RetrieveTokenByAccessTokenHandler: RetrieveTokenByAccessToken,
                    RetrieveTokenByRefreshTokenHandler: RetrieveTokenByRefreshToken
                })
        return self._query_bus



from flask import url_for, request
from flask_restplus import Namespace, Resource

import strongr.restdomain.model.gateways
from strongr.restdomain.model.oauth2 import Client

ns = Namespace('oauth', description='Operations related to oauth2 login')

@ns.route('/revoke', methods=['POST'])
class Revoke(Resource):
    def post(self):
        auth_server = strongr.restdomain.model.gateways.Gateways.auth_server()
        return auth_server.create_revocation_response()

@ns.route('/token', methods=['POST'])
class Token(Resource):
    def post(self):
        auth_server = strongr.restdomain.model.gateways.Gateways.auth_server()
        return auth_server.create_token_response()

# @ns.route('/authorize')
# class Authorize(Resource):
#     def get(self):
#         auth_server = strongr.restdomain.model.gateways.Gateways.auth_server()
#         # Login is required since we need to know the current resource owner.
#         # It can be done with a redirection to the login page, or a login
#         # form on this authorization page.
#         if request.method == 'GET':
#             grant = auth_server.validate_authorization_request()
#             #return render_template(
#             #    'authorize.html',
#             #    grant=grant,
#             #    user=current_user,
#             #)
#         confirmed = request.form['confirm']
#         if confirmed:
#             # granted by resource owner
#             return auth_server.create_authorization_response()
#         # denied by resource owner
#         return auth_server.create_authorization_response(None)
#
#     def post(self):
#         auth_server = strongr.restdomain.model.gateways.Gateways.auth_server()
#         # Login is required since we need to know the current resource owner.
#         # It can be done with a redirection to the login page, or a login
#         # form on this authorization page.
#         #if request.method == 'GET':
#             #grant = auth_server.validate_authorization_request()
#             # return render_template(
#             #    'authorize.html',
#             #    grant=grant,
#             #    user=current_user,
#             # )
#         #confirmed = request.form['confirm']
#         #if confirmed:
#             # granted by resource owner
#         return auth_server.create_authorization_response('1')
#         # denied by resource owner
#         #return auth_server.create_authorization_response(None)



import strongr.core

from sqlalchemy import func

from strongr.schedulerdomain.model import Job, JobState


class RequestResourcesRequiredHandler(object):
    def __call__(self, query):
        db = strongr.core.gateways.Gateways.sqlalchemy_session()

        results = db.query(func.count(Job.cores).label('cores'), func.count(Job.ram).label('ram')).filter(Job.state == JobState.ENQUEUED).all()

        if results:
            return {'cores': results[0][0], 'ram': results[0][1]}
        return None



from .requestscheduledjobs import RequestScheduledJobs
from .requestjobinfo import RequestJobInfo
from .findnodewithavailableresources import FindNodeWithAvailableResources
from .requestfinishedjobs import RequestFinishedJobs
from .requestresourcesrequired import RequestResourcesRequired
from .requestvmsbystate import RequestVmsByState



import dependency_injector.containers as containers
import dependency_injector.providers as providers

from strongr.restdomain.service import Oauth2Service
from strongr.restdomain.factory.oauth2 import CommandFactory as Oauth2CommandFactory,\
                                                QueryFactory as Oauth2QueryFactory


class RestDomain(containers.DeclarativeContainer):
    """IoC container"""
    oauth2CommandFactory = providers.Singleton(Oauth2CommandFactory)
    oauth2QueryFactory = providers.Singleton(Oauth2QueryFactory)
    oauth2Service = providers.Singleton(Oauth2Service)



from strongr.core.stats.abstractstats import AbstractStats


class NullContextManager:
    def __enter__(self):
        pass

    def __exit__(self, type, value, traceback):
        pass


class NullDriver(AbstractStats):
    def __init__(self, *args, **kwargs):
        pass

    def decr(self, namespace, amount, rate=None):
        pass

    def gauge(self, namespace, amount, rate=False, delta=False):
        pass

    def set(self, namespace, arr):
        pass

    def time(self, namespace):
        return NullContextManager()

    def timing(self, namespace, timems):
        pass

    def incr(self, namespace, amount, rate=None):
        pass



import dependency_injector.containers as containers
import dependency_injector.providers as providers

import strongr.core

from strongr.schedulerdomain.model.scalingdrivers.nullscaler import NullScaler
from strongr.schedulerdomain.model.scalingdrivers.simplescaler import SimpleScaler
from strongr.schedulerdomain.model.scalingdrivers.surfhpccloudscaler import SurfHpcScaler


class ScalingDriver(containers.DeclarativeContainer):
    """IoC container of service providers."""
    _scalingdrivers = providers.Object({
        'simplescaler': SimpleScaler,
        'nullscaler':  NullScaler,
        'surfsarahpccloud': SurfHpcScaler
    })

    scaling_driver = providers.Singleton(_scalingdrivers()[strongr.core.Core.config().schedulerdomain.scalingdriver.lower()], config=dict(strongr.core.Core.config().schedulerdomain.as_dict()[strongr.core.Core.config().schedulerdomain.scalingdriver]) if strongr.core.Core.config().schedulerdomain.scalingdriver in strongr.core.Core.config().schedulerdomain.as_dict().keys() else {})



from sqlalchemy import func

import strongr.core
from strongr.core.gateways import Gateways
from strongr.schedulerdomain.model import Vm, VmState, Job, JobState


class LogStatsHandler(object):
    def __call__(self, command):
        if strongr.core.Core.config().stats.driver.lower() == 'null':
            return # skip collecting stats, nulldriver is selected

        session = Gateways.sqlalchemy_session()
        stats = Gateways.stats()

        # number of VM's, cores, ram
        active_vms = session.query(
            func.count(Vm.vm_id),
            func.sum(Vm.cores),
            func.sum(Vm.ram)
        ).filter(Vm.state.in_([VmState.READY, VmState.MARKED_FOR_DEATH])).all()

        if len(active_vms) > 0:
            active_vms = active_vms[0] # only 1 row is returned
            namespace = 'stats.vms.active.'
            stats.gauge(namespace + 'count', active_vms[0] if active_vms[0] is not None else 0)
            stats.gauge(namespace + 'cores', active_vms[1] if active_vms[1] is not None else 0)
            stats.gauge(namespace + 'ram', active_vms[2] if active_vms[2] is not None else 0)

        # number of active jobs, cores used, ram used
        active_jobs = session.query(
            func.count(Job.job_id),
            func.sum(Job.cores),
            func.sum(Job.ram)
        ).filter(Job.state == JobState.RUNNING).all()

        if len(active_jobs) > 0:
            active_jobs = active_jobs[0] # only 1 row is returned
            namespace = 'stats.jobs.running.'
            stats.gauge(namespace + 'count', active_jobs[0] if active_jobs[0] is not None else 0)
            stats.gauge(namespace + 'cores', active_jobs[1] if active_jobs[1] is not None else 0)
            stats.gauge(namespace + 'ram', active_jobs[2] if active_jobs[2] is not None else 0)

        # number of enqueued jobs, cores needed, ram needed
        enqueued_jobs = session.query(
            func.count(Job.job_id),
            func.sum(Job.cores),
            func.sum(Job.ram)
        ).filter(Job.state == JobState.ENQUEUED).all()

        if len(enqueued_jobs) > 0:
            enqueued_jobs = enqueued_jobs[0]  # only 1 row is returned
            namespace = 'stats.jobs.enqueued.'
            stats.gauge(namespace + 'count', enqueued_jobs[0] if enqueued_jobs[0] is not None else 0)
            stats.gauge(namespace + 'cores', enqueued_jobs[1] if enqueued_jobs[1] is not None else 0)
            stats.gauge(namespace + 'ram', enqueued_jobs[2] if enqueued_jobs[2] is not None else 0)

        # number of failed jobs, cores needed, ram needed
        failed_jobs = session.query(
            func.count(Job.job_id)
        ).filter(Job.state == JobState.FAILED).all()

        if len(failed_jobs) > 0:
            failed_jobs = failed_jobs[0]  # only 1 row is returned
            namespace = 'stats.jobs.failed.'
            stats.gauge(namespace + 'count', failed_jobs[0] if failed_jobs[0] is not None else 0)



from functools import wraps
import strongr.restdomain.model.gateways

# oauth2 lib does not support namespaces so we need a hack
# https://github.com/lepture/flask-oauthlib/issues/180
def namespace_require_oauth(*scopes):
    def wrapper(f):
        @wraps(f)
        def check_oauth(*args, **kwargs):
            return strongr.restdomain.model.gateways.Gateways.require_oauth()(*scopes)(f)(*args, **kwargs)

        return check_oauth
    return wrapper



from .schedulejob import ScheduleJob
from .runenqueuedjobs import RunEnqueuedJobs
from .startjobonvm import StartJobOnVm
from .checkjobsrunning import CheckJobsRunning
from .ensureminamountofnodes import EnsureMinAmountOfNodes
from .scaleout import ScaleOut
from .jobfinished import JobFinished
from .vmcreated import VmCreated
from .vmready import VmReady
from .vmdestroyed import VmDestroyed
from .vmnew import VmNew
from .checkscaling import CheckScaling
from .cleanupnodes import CleanupNodes
from .scalein import ScaleIn
from .logstats import LogStats
from .cleanupoldjobs import CleanupOldJobs



import uuid

import strongr.core
import logging

class EnsureMinAmountOfNodesHandler(object):
    def __call__(self, command):
        config = strongr.core.Core.config()
        templates = config.schedulerdomain.simplescaler.templates.as_dict()
        logger = logging.getLogger('schedulerdomain.' + self.__class__.__name__)

        cloudQueryBus = strongr.core.domain.clouddomain.CloudDomain.cloudService().getQueryBus()
        cloudQueryFactory = strongr.core.domain.clouddomain.CloudDomain.queryFactory()

        machines = cloudQueryBus.handle(cloudQueryFactory.newListDeployedVms())

        template_counters = {}
        machines_needed = {}
        for machine in machines:
            for template_name in templates:
                if machine.startswith(template_name + '-'):
                    if template_name in template_counters:
                        template_counters[template_name] += 1
                    else:
                        template_counters[template_name] = 1
                    break

        for template_name in templates:
            if template_name in template_counters and template_counters[template_name] < templates[template_name]['spawned-min']:
                machines_needed[template_name] = templates[template_name]['spawned-min'] - template_counters[template_name]
            else:
                machines_needed[template_name] = templates[template_name]['spawned-min']


        if len(machines_needed) > 0:
            logger.info('Machines needed: {}'.format(machines_needed))
            for machine in machines_needed:
                cloudService = strongr.core.domain.clouddomain.CloudDomain.cloudService()
                commandFactory = strongr.core.domain.clouddomain.CloudDomain.commandFactory()
                cloudProviderName = config.clouddomain.driver
                profile = getattr(config.clouddomain, cloudProviderName).default_profile if 'profile' not in templates[machine] else templates[machine]['profile']
                names = []
                for i in range(0, machines_needed[machine]):
                    names.append(machine + '-' + str(uuid.uuid4()))

                deployVmsCommand = commandFactory.newDeployVms(names=names, profile=profile, cores=templates[machine]['cores'], ram=templates[machine]['ram'])

                commandBus = cloudService.getCommandBus()

                logger.info('Deploying VMs {0} cores={1} ram={2}GiB'.format(names, templates[machine]['cores'], templates[machine]['ram']))

                commandBus.handle(deployVmsCommand)
        else:
            logger.info('No aditional machines needed!')



from strongr.clouddomain.model.gateways import Gateways
from .abstractcloudservice import AbstractCloudService

from strongr.clouddomain.handler.impl.cloud.mockcloud import ListDeployedVmsHandler, RunJobHandler, DeployVmsHandler

import subprocess
import threading
import strongr.core

class MockCloud(AbstractCloudService):
    def __init__(self, *args, **kwargs):
        super(MockCloud, self).__init__(*args, **kwargs)
        self._check_docker()
        thread = threading.Thread(target=self._publish_localhost)  # run in separate thread so it doesn't block strongr
        thread.start()

    def _publish_localhost(self):
        import time
        import psutil

        time.sleep(5)

        inter_domain_event_factory = Gateways.inter_domain_event_factory()
        inter_domain_events_publisher = strongr.core.Core.inter_domain_events_publisher()

        vmnew_event = inter_domain_event_factory.newVmNewEvent('localhost', psutil.cpu_count(logical=True), psutil.virtual_memory().total / 1024 / 1024 / 1024)
        inter_domain_events_publisher.publish(vmnew_event)

        time.sleep(1)
        vmcreated_event = inter_domain_event_factory.newVmCreatedEvent('localhost')
        inter_domain_events_publisher.publish(vmcreated_event)

        time.sleep(1)
        vmready_event = inter_domain_event_factory.newVmReadyEvent('localhost')
        inter_domain_events_publisher.publish(vmready_event)


    def _check_docker(self):
        ret_code = subprocess.call('docker ps', shell=True)
        if ret_code != 0:
            raise Exception("Can't access docker sock. Is docker installed? Do you have sufficient privileges?")

    def get_command_handlers(self):
        return [RunJobHandler, DeployVmsHandler]

    def get_query_handlers(self):
        return [ListDeployedVmsHandler]



import dependency_injector.containers as containers
import dependency_injector.providers as providers
from redis import Redis

from strongr.core.cache import get_cache
from strongr.core.lock.lockfactory import get_lock
import strongr.core

from sqlalchemy.orm import sessionmaker
from sqlalchemy import engine_from_config
from sqlalchemy.ext.declarative import declarative_base

from strongr.core.stats.statsddriver import StatsDDriver
from strongr.core.stats.nulldriver import NullDriver

class Gateways(containers.DeclarativeContainer):
    """IoC container of gateway components."""

    _stats_drivers = providers.Object({
        "statsd": StatsDDriver,
        "null": NullDriver
    })

    cache = providers.Singleton(get_cache)
    lock = providers.Factory(get_lock)

    redis = providers.Singleton(Redis.from_url, url=(strongr.core.Core.config().redis.url if hasattr(strongr.core.Core.config(),'redis') else ''))

    sqlalchemy_engine = providers.ThreadLocalSingleton(engine_from_config, configuration=strongr.core.Core.config().db.engine.as_dict(), prefix='') # construct engine from config
    sqlalchemy_session = providers.ThreadLocalSingleton(sessionmaker(bind=sqlalchemy_engine()))
    sqlalchemy_base = providers.Singleton(declarative_base)

    stats = providers.Singleton(_stats_drivers()[strongr.core.Core.config().stats.driver], config=(strongr.core.Core.config().stats.config.as_dict() if hasattr(strongr.core.Core.config().stats, 'config') else {}))



from strongr.clouddomain.query import ListDeployedVms, RequestJidStatus

from strongr.core.exception import InvalidParameterException

class QueryFactory:
    def newListDeployedVms(self):
        """ Generates a new ListDeployedVms query

        :returns: A ListDeployedVms query object
        :rtype: ListDeployedVms
        """
        return ListDeployedVms()

    def newRequestJidStatus(self, jid):
        """ Generates a new RequestJidStatusquery

        :returns: A RequestJidStatus query object
        :rtype: RequestJidStatus
        """
        return RequestJidStatus(jid)




import strongr.core

import strongr.core.domain.schedulerdomain
import strongr.core.domain.clouddomain
import strongr.core.gateways

from sqlalchemy import and_
from datetime import datetime, timedelta

from strongr.schedulerdomain.model import Job, JobState


class CheckJobsRunningHandler:
    def __call__(self, command):
        session = strongr.core.gateways.Gateways.sqlalchemy_session()

        deadline = datetime.utcnow() - timedelta(minutes=15)

        for job in session.query(Job).filter(and_(Job.state == JobState.RUNNING, Job.state_date < deadline)).all():
            cloudQueryBus = strongr.core.domain.clouddomain.CloudDomain.cloudService().getQueryBus()
            cloudQueryFactory = strongr.core.domain.clouddomain.CloudDomain.queryFactory()

            status = cloudQueryBus.handle(cloudQueryFactory.newRequestJidStatus(job.job_id))

            if status is None or not status:
                # job not finished yet
                continue

            session.query(Job).filter(Job.job_id == job.job_id).update({Job.stdout: status[list(status.keys())[0]], Job.return_code: 0, Job.state: JobState.FINISHED}, synchronize_session='evaluate')



import uuid

from celery.signals import worker_process_init

from strongr.core.domain.schedulerdomain import SchedulerDomain
from .wrapper import Command

import strongr.core
from celery import Celery, signals


class RunWorkerCommand(Command):
    """
    Runs a strongr worker.

    worker:run
        {--p|profile=elasticworker : The worker profile to be used}
    """
    def handle(self):
        config = strongr.core.Core.config()
        broker = config.celery.broker
        backend = config.celery.backend

        celery = Celery('strongr', broker=broker, backend=backend)

        scheduler = SchedulerDomain.schedulerService()
        scheduler.getCommandBus() # we need to initiate this as it self-registers it's commands to celery
        scheduler.getQueryBus() # we need to initiate this as it self-registers it's queries to celery

        domains = getattr(config.celery.workers, self.option('profile')).as_dict()

        commands = []
        for domain in domains:
            for command in domains[domain]:
                commands.append('{}.{}'.format(domain, command))

        for command in commands:
            strongr.core.Core.command_router().enable_worker_route_for_command(celery, command)

        workername = '%h@{}'.format(uuid.uuid4())

        argv = [
            'worker',
            '--loglevel=DEBUG',
            '-Q=' + ','.join(commands),
            #'--pool=solo',
            #'--concurrency=1',
            #'--max-tasks-per-child=1',
            '--loglevel=INFO',
            '-n {}'.format(workername)
        ]

        # crude fix for celery daemonized threads causing salt to fail
        # https://github.com/celery/celery/issues/1709
        @worker_process_init.connect
        def fix_multiprocessing(**kwargs):
            # don't be a daemon, so we can create new subprocesses
            from multiprocessing import current_process
            current_process().daemon = False

        @signals.setup_logging.connect
        def setup_celery_logging(**kwargs):
            pass

        celery.worker_main(argv)



from .retrieveclient import RetrieveClient
from .retrievegrant import RetrieveGrant
from .retrievetokenbyaccesstoken import RetrieveTokenByAccessToken
from .retrievetokenbyrefreshtoken import RetrieveTokenByRefreshToken



import filelock
import os

class FileLock(filelock.FileLock):
    def __init__(self, name, path, *args, **kwargs):
        if not os.path.isdir(path):
            os.makedirs(path)

        super(FileLock, self).__init__(lock_file=os.path.join(path, name), *args, **kwargs)

    def exists(self):
        # unfortunately there is no proper function in the filelock lib to check this
        try:
            self.acquire(timeout=0.001,poll_intervall=0.001)
            self.release()
        except:
            return True
        return False

    def __enter__(self):
        self.acquire()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.release()



from setuptools import setup, find_packages

setup(
    name='strongr',
    version='1.1',
    description='An elastic cloud command runner',
    author='Thomas Phil',
    author_email='thomas@tphil.nl',
    packages=find_packages(),  # auto detect packages
    install_requires=[
        'salt==2018.3.1'
        'jsonpickle==0.9.5',
        'cleo==0.6.1',
        'Flask==0.12.2',
        'dependency_injector==3.6.1',
        'filelock==2.0.12',
        'flask_restplus==0.10.1',
        'Flask_OAuthlib==0.9.4',
        'cmndr==1.0.5',
        'gunicorn==19.7.1',
        'SQLAlchemy==1.1.14',
        'PyYAML==3.12',
        'psutil==5.4.3',
        'celery==4.1.0',
        'redis==2.10.6',
        'statsd==3.2.2',
        'enum34==1.1.6',
        'schedule',
        'authlib==0.4.1',
        'mysqlclient',
        'tornado==4.5.2',
    ], #external packages as dependencies
    scripts=[
        'bin/strongr'
    ]
)



from .wrapper import Command

import unittest

class RunTestsCommand(Command):
    """
    Runs the unit tests

    unittests:run
    """
    def handle(self):
        testsuite = unittest.TestLoader().discover('.')
        unittest.TextTestRunner(verbosity=1).run(testsuite)



import celery
import strongr.core
import jsonpickle

class Envelope(celery.Task):
    def patch(self, celery_app, name, ignore_result):
        self.name = name
        self.app = celery_app
        # register ourselves with celery
        # this way celery knows about this command
        celery_app.tasks.register(self)
        #print('{} ignore result: {}'.format(name, ignore_result))
        self._ignore_result = ignore_result

    def run(self, command, *args, **kwargs):
        core = strongr.core.Core()

        command = jsonpickle.decode(command)

        result = core.command_router().handle_command(command)
        # ignore return value if flag is true
        # workaround from https://github.com/celery/celery/issues/1904
        if self._ignore_result:
            raise celery.exceptions.Ignore()
        return result



from strongr.core import Core
from strongr.core.domain.clouddomain import CloudDomain
from .wrapper import Command

class DestroySingleCommand(Command):
    """
    Destroys a VM in the cloud.

    destroy:single
        {machine : The name of the VM to be destroyed}
    """
    def handle(self):
        cloudService = CloudDomain.cloudService()
        commandFactory = CloudDomain.commandFactory()

        machine = self.argument('machine')
        destroyVmsCommand = commandFactory.newDestroyVms(names=[machine])

        commandBus = cloudService.getCommandBus()

        self.info('Destroying VM {0}'.format(machine))

        commandBus.handle(destroyVmsCommand)



import sys

if sys.version_info > (3, 0):
    import configparser as ConfigParser
else:
    import ConfigParser
import os.path

class IniLoader():
    def getConfig(self, environment):
        configLocations = ['/etc/strongr/config.ini', os.path.expanduser('~/.strongr/config.ini'), 'config.ini']
        output = {}

        return output # return nothing for now, below code does not work anymore

        for configLocation in configLocations:
            if os.path.isfile(configLocation):
                config = ConfigParser.ConfigParser()
                config.read(configLocation)
                for key in config.defaults():
                    output[key] = config.defaults()[key]
                for key in config.options(environment):
                    output[key] = config.get(environment, key)

        return output




import dependency_injector.containers as containers
import dependency_injector.providers as providers
from flask import Flask

from werkzeug.security import generate_password_hash, check_password_hash

from authlib.flask.oauth2 import AuthorizationServer, ResourceProtector
from strongr.restdomain.api.apiv1 import blueprint as apiv1
from strongr.restdomain.model.oauth2 import Token

from strongr.restdomain.model.oauth2.client import Client

import strongr.core
from strongr.restdomain.model.oauth2.endpoints.revocationendpoint import RevocationEndpoint
from strongr.restdomain.model.oauth2.grants.authorizationcodegrant import AuthorizationCode
from strongr.restdomain.model.oauth2.grants.clientcredentialsgrant import ClientCredentialsGrant
from strongr.restdomain.model.oauth2.grants.passwordgrant import PasswordGrant

class Gateways(containers.DeclarativeContainer):
    """IoC container of gateway objects."""
    _blueprints = providers.Object([apiv1])

    def _factor_app(name, blueprints):
        app = Flask(__name__)

        config = strongr.core.Core.config()
        backend = config.restdomain.backend.strip().lower()

        # the oauth2 lib can not work with templates,
        # this hack was proposed as a temp fix on the
        # libraries github. Use this for now, we
        # should refactor this later.
        # https://github.com/lepture/flask-oauthlib/issues/180
        #from strongr.restdomain.api.oauth2 import bind_oauth2
        #oauth2 = bind_oauth2(app)
        #app.oauth2 = oauth2

        for blueprint in blueprints:
            app.register_blueprint(blueprint)

        #auth_server = AuthorizationServer(Client, app)

        #auth_server.register_grant_endpoint(AuthorizationCode)
        #auth_server.register_grant_endpoint(PasswordGrant)
        #auth_server.register_grant_endpoint(ClientCredentialsGrant)

        #auth_server.register_revoke_token_endpoint(RevocationEndpoint)

        #Gateways.auth_server.override(auth_server)

        if backend == 'flask':
            flask_config = config.restdomain.flask.as_dict() if hasattr(config, 'restdomain') and hasattr(config.restdomain, 'flask') else {}
            # monkey patch run method so that config is grabbed from config file
            setattr(app, '_run_original', app.run)
            app.run = lambda self=app: self._run_original(**flask_config)
            return app
        elif backend == 'gunicorn':
            from gunicorn.app.base import BaseApplication

            # put WSGIServer class here for now
            # this should be refactored later
            class WSGIServer(BaseApplication):
                def __init__(self, app):
                    self.application = app
                    super(WSGIServer, self).__init__("%(prog)s [OPTIONS]")

                def load_config(self):
                    for key in config.restdomain.gunicorn.as_dict():
                        self.cfg.set(key, config.restdomain.gunicorn.as_dict()[key])

                def load(self):
                    return self.application

            return WSGIServer(app)

    @classmethod
    def _query_token(self, access_token):
        from authlib.flask.oauth2 import current_token
        import strongr.core.gateways

        return strongr.core.gateways.Gateways.sqlalchemy_session().query(Token).filter_by(access_token=access_token).first()


    def _factor_require_oauth():
        return ResourceProtector(Gateways._query_token)

    flask_app = providers.Singleton(_factor_app, 'StrongRRestServer', _blueprints())
    auth_server = providers.Configuration('auth_server') # initialized by _factor_app factory
    #current_user = providers.Factory(_factor_current_user)
    require_oauth = providers.Factory(_factor_require_oauth)

    generate_password_hash = providers.Factory(generate_password_hash, method='pbkdf2:sha256:80000', salt_length=8)
    check_password_hash = providers.Factory(check_password_hash)



from strongr.schedulerdomain.model.scalingdrivers import ScalingDriver


class ScaleOutHandler(object):
    def __call__(self, command):
        # forward to configured scaleout driver
        ScalingDriver.scaling_driver().scaleout(command)
